from libcpp cimport bool
from Feature cimport *
from MRMFeature cimport *
from UniqueIdInterface cimport *
from ProteinIdentification cimport *
from PeptideIdentification cimport *
from DataProcessing cimport *
from MetaInfoInterface cimport *
from DocumentIdentifier cimport *
from RangeManager cimport *
from MSExperiment cimport *

# this class has addons, see the ./addons folder

cdef extern from "<OpenMS/KERNEL/FeatureMap.h>" namespace "OpenMS":

    cdef cppclass FeatureMap(UniqueIdInterface, DocumentIdentifier, RangeManager2, MetaInfoInterface):

        # wrap-inherits:
        #   UniqueIdInterface
        #   DocumentIdentifier
        #   RangeManager2
        #   MetaInfoInterface
        #
        # wrap-instances:
        #   FeatureMap := FeatureMap
        #
        # wrap-doc:
        #   A container for features.
        #   -----
        #   A feature map is a container holding features, which represent
        #   chemical entities (peptides, proteins, small molecules etc.) found
        #   in an LC-MS/MS experiment.
        #   -----
        #   This class supports direct iteration in Python.

        FeatureMap() nogil except +
        FeatureMap(FeatureMap &) nogil except +

        bool operator==(FeatureMap) nogil except +
        bool operator!=(FeatureMap) nogil except +

        int size()  nogil except +
        Feature & operator[](int)      nogil except + #wrap-upper-limit:size()
        void push_back(Feature spec) nogil except +
        void push_back(MRMFeature spec) nogil except +

        void sortByIntensity() nogil except +
        void sortByIntensity(bool reverse) nogil except +
        void sortByPosition() nogil except +
        void sortByRT() nogil except +
        void sortByMZ() nogil except +
        void sortByOverallQuality() nogil except +

        void swap(FeatureMap &) nogil except +
        void swapFeaturesOnly(FeatureMap swapfrom) nogil except +
        void clear() nogil except +
        void clear(bool clear_meta_data) nogil except +

        FeatureMap operator+(FeatureMap) nogil except +
        FeatureMap iadd(FeatureMap) nogil except + # wrap-as:operator+=

        void updateRanges() nogil except +

        libcpp_vector[ProteinIdentification] getProteinIdentifications() nogil except+
        void setProteinIdentifications(libcpp_vector[ProteinIdentification]) nogil except+

        libcpp_vector[PeptideIdentification] getUnassignedPeptideIdentifications() nogil except+
        void setUnassignedPeptideIdentifications(libcpp_vector[PeptideIdentification]) nogil except+

        Size applyMemberFunction(Size(* fun)()) nogil except +# wrap-ignore

        libcpp_vector[DataProcessing] getDataProcessing() nogil except +
        void setDataProcessing(libcpp_vector[DataProcessing])   nogil except +

        void setPrimaryMSRunPath(StringList& s) nogil except +
        void setPrimaryMSRunPath(StringList& s, MSExperiment& e) nogil except +
        void getPrimaryMSRunPath(StringList& toFill) nogil except +

        libcpp_vector[Feature].iterator begin() nogil except +    # wrap-iter-begin:__iter__(Feature)
        libcpp_vector[Feature].iterator end()   nogil except +    # wrap-iter-end:__iter__(Feature)




from Types cimport *
# from MathFunctions cimport *
from ContinuousWaveletTransform cimport *

cdef extern from "<OpenMS/TRANSFORMATIONS/RAW2PEAK/ContinuousWaveletTransformNumIntegration.h>" namespace "OpenMS":
    
    cdef cppclass ContinuousWaveletTransformNumIntegration(ContinuousWaveletTransform) :
        # wrap-inherits:
        #  ContinuousWaveletTransform
        ContinuousWaveletTransformNumIntegration() nogil except +
        ContinuousWaveletTransformNumIntegration(ContinuousWaveletTransformNumIntegration) nogil except + #wrap-ignore

        # TODO iterator
        # TEMPLATE # void transform(InputPeakIterator begin_input, InputPeakIterator end_input, float resolution, unsigned int zeros) nogil except +

        # parent class
        #void init(double scale, double spacing) nogil except +




from libcpp.vector cimport vector as libcpp_vector
from libcpp cimport bool
from DPosition cimport *
from String cimport *

cdef extern from "<OpenMS/MATH/STATISTICS/GaussFitter.h>" namespace "OpenMS::Math":

    cdef cppclass GaussFitter:

        GaussFitter() nogil except +
        GaussFitter(GaussFitter) nogil except +   # wrap-ignore

        # sets the initial parameters used by the fit method as inital guess for the gaussian
        void setInitialParameters(GaussFitResult & result) nogil except +

        #   @brief Fits a gaussian distribution to the given data points
        #   @param points the data points used for the gaussian fitting
        #   @exception Exception::UnableToFit is thrown if fitting cannot be performed
        GaussFitResult fit(libcpp_vector[DPosition2] points) nogil except +

cdef extern from "<OpenMS/MATH/STATISTICS/GaussFitter.h>" namespace "OpenMS::Math::GaussFitter":

    cdef cppclass GaussFitResult:

        GaussFitResult() nogil except +
        GaussFitResult(double, double, double) nogil except +
        GaussFitResult(GaussFitResult) nogil except +   # wrap-ignore

        double eval(double) nogil except +

        # parameter A of gaussian distribution (amplitude)
        double A
        # parameter x0 of gaussian distribution (left/right shift)
        double x0
        # parameter sigma of gaussian distribution (width)
        double sigma



from Types cimport *
from Types cimport *
from DateTime cimport *
from String cimport *

cdef extern from "<OpenMS/FORMAT/HANDLERS/XMLHandler.h>" namespace "OpenMS::Internal":
    
    # inherits from xercesc DefaultHandler -> no wrapping of it xercesc classes
    cdef cppclass XMLHandler:
        XMLHandler() nogil except + #wrap-ignore
        XMLHandler(XMLHandler) nogil except + #wrap-ignore

        # NAMESPACE # void fatalError(xercesc::SAXParseException & exception) nogil except +
        # NAMESPACE # void error(xercesc::SAXParseException & exception) nogil except +
        # NAMESPACE # void warning(xercesc::SAXParseException & exception) nogil except +
        XMLHandler(const String & filename, const String & version) nogil except +
        void reset() nogil except +
        # TODO cdash might parse out "fatalError" statements and interpret them
        # as compilation failure...
        # void fatalError(ActionMode mode, const String & msg, UInt line, UInt column) nogil except +
        void error(ActionMode mode, const String & msg, UInt line, UInt column) nogil except +
        void warning(ActionMode mode, const String & msg, UInt line, UInt column) nogil except +
        # POINTER # void characters(XMLCh *chars, XMLSize_t length) nogil except +
        # NAMESPACE # # POINTER # void startElement(XMLCh *uri, XMLCh *localname, XMLCh *qname, xercesc::Attributes & attrs) nogil except +
        # POINTER # void endElement(XMLCh *uri, XMLCh *localname, XMLCh *qname) nogil except +
        # NAMESPACE # void writeTo(std::ostream & ) nogil except +
        String errorString() nogil except +

cdef extern from "<OpenMS/FORMAT/HANDLERS/XMLHandler.h>" namespace "OpenMS::Internal::XMLHandler":
    cdef enum ActionMode "OpenMS::Internal::XMLHandler::ActionMode":
        #wrap-attach:
        #    XMLHandler
        LOAD
        STORE




from Types cimport *
from DefaultParamHandler cimport *
from BaseFeature cimport *
from ProgressLogger cimport *
from String cimport *
from StringList cimport *
from MSExperiment cimport *
from Map cimport *
from FeatureMapping cimport *

cdef extern from "<OpenMS/ANALYSIS/ID/SiriusMSConverter.h>" namespace "OpenMS":
    
    cdef cppclass SiriusMSFile "OpenMS::SiriusMSFile":

        SiriusMSFile() nogil except +
        SiriusMSFile(SiriusMSFile) nogil except + #wrap-ignore

        void store(MSExperiment& spectra,
                   String& msfile,
                   FeatureMapping_FeatureToMs2Indices& feature_ms2_spectra_map,
                   bool& feature_only,
                   int& isotope_pattern_iterations,
                   bool no_mt_info,
                   libcpp_vector[ SiriusMSFile_CompoundInfo ] v_cmpinfo) nogil except +

    cdef cppclass SiriusMSFile_CompoundInfo "OpenMS::SiriusMSFile::CompoundInfo":

        SiriusMSFile_CompoundInfo() nogil except +
        SiriusMSFile_CompoundInfo(SiriusMSFile_CompoundInfo) nogil except + #wrap-ignore

    cdef cppclass SiriusMSFile_AccessionInfo "OpenMS::SiriusMSFile::AccessionInfo":

        SiriusMSFile_AccessionInfo() nogil except +
        SiriusMSFile_AccessionInfo(SiriusMSFile_AccessionInfo) nogil except + #wrap-ignore


from Types cimport *
from String cimport *
from libcpp cimport bool
from libcpp.vector cimport vector as libcpp_vector
from libcpp.set cimport set as libcpp_set

from MSExperiment cimport *
from FeatureMap cimport *
from Precursor cimport *


cdef extern from "<OpenMS/FILTERING/CALIBRATION/PrecursorCorrection.h>" namespace "OpenMS":

    cdef cppclass PrecursorCorrection:

        void getPrecursors(MSExperiment & exp,
                           libcpp_vector[ Precursor ] & precursors,
                           libcpp_vector[ double ] & precursors_rt,
                           libcpp_vector[ size_t ] & precursor_scan_index) nogil except +

        void writeHist(String & out_csv,
                       libcpp_vector[ double ] & delta_mzs,
                       libcpp_vector[ double ] & mzs,
                       libcpp_vector[ double ] & rts) nogil except +

        libcpp_set[ size_t ] correctToNearestMS1Peak(MSExperiment & exp,
                                                     double mz_tolerance,
                                                     bool ppm,
                                                     libcpp_vector[ double ] & delta_mzs,
                                                     libcpp_vector[ double ] & mzs,
                                                     libcpp_vector[ double ] & rts) nogil except +

        libcpp_set[ size_t ] correctToHighestIntensityMS1Peak(MSExperiment & exp,
                                                              double mz_tolerance,
                                                              bool ppm,
                                                              libcpp_vector[ double ] & delta_mzs,
                                                              libcpp_vector[ double ] & mzs,
                                                              libcpp_vector[ double ] & rts) nogil except +
        
        libcpp_set[ size_t ] correctToNearestFeature(FeatureMap & features,
                                                     MSExperiment & exp,
                                                     double rt_tolerance_s,
                                                     double mz_tolerance,
                                                     bool ppm,
                                                     bool believe_charge,
                                                     bool keep_original,
                                                     bool all_matching_features,
                                                     int max_trace,
                                                     int debug_level) nogil except +




from libcpp cimport bool
from Types cimport *

# this class has addons, see the ./addons folder

cdef extern from "<OpenMS/DATASTRUCTURES/DPosition.h>" namespace "OpenMS":

    # this is the way to declare a class with a int template parameter.

    # it is a bit laborous as one has to declare all methods twice,
    # but deriving from a common base class and giving an extra string
    # with explicit c++ type declarion does not work !

    # as an alternative one could use a similar ctypedef, but these can
    # not be renamed when cipmorting, which might cause trouble !

    cdef cppclass DPosition1 "OpenMS::DPosition<1> ":
        # wrap-ignore
        DPosition1()  nogil except +
        DPosition1(double)  nogil except +
        DPosition1(DPosition1)  nogil except +

        double & operator[](Size index) nogil except +
        bool operator==(DPosition1) nogil except +
        bool operator!=(DPosition1) nogil except +

        bool operator<(DPosition1) nogil except +
        bool operator<=(DPosition1) nogil except +

        bool operator>(DPosition1) nogil except +
        bool operator>=(DPosition1) nogil except +

        void clear() nogil except +
        Size size() nogil except +

    cdef cppclass DPosition2 "OpenMS::DPosition<2> ":
        # wrap-ignore
        DPosition2()  nogil except +
        DPosition2(DPosition2)  nogil except +
        DPosition2(double)  nogil except +
        DPosition2(double, double)  nogil except +

        double & operator[](Size index) nogil except +
        bool operator==(DPosition2) nogil except +
        bool operator!=(DPosition2) nogil except +

        bool operator<(DPosition2) nogil except +
        bool operator<=(DPosition2) nogil except +

        bool operator>(DPosition2) nogil except +
        bool operator>=(DPosition2) nogil except +

        void clear() nogil except +
        Size size() nogil except +



from libcpp.vector cimport vector as libcpp_vector
from libcpp.string cimport string as libcpp_string
from FeatureMap cimport *
from TargetedExperiment cimport *
from TransformationDescription cimport *

cdef extern from "<OpenMS/ANALYSIS/OPENSWATH/ConfidenceScoring.h>" namespace "OpenMS":

    cdef cppclass ConfidenceScoring:

        ConfidenceScoring() nogil except +
        ConfidenceScoring(ConfidenceScoring) nogil except +

        void initialize(TargetedExperiment & targeted, Size n_decoys, Size n_transitions, TransformationDescription trafo) nogil except +
        void initializeGlm(double intercept, double rt_coef, double int_coef) nogil except +
        void scoreMap(FeatureMap & map) nogil except +




from Types cimport *

cdef extern from "<OpenMS/MATH/MISC/BSpline2d.h>" namespace "OpenMS":
    
    cdef cppclass BSpline2d:

        BSpline2d(libcpp_vector[double] x, libcpp_vector[double] y, 
                  double wave_length, BoundaryCondition boundary_condition, 
                  Size num_nodes) nogil except +

        bool solve(libcpp_vector[double] y) nogil except +

        double eval(double x) nogil except +

        double derivative(double x) nogil except +

        bool ok() nogil except +

        void debug(bool enable) nogil except +

        double derivatives(double x, unsigned order) nogil except +

cdef extern from "<OpenMS/MATH/MISC/BSpline2d.h>" namespace "OpenMS::BSpline2d":

    cdef enum BoundaryCondition:
        BC_ZERO_ENDPOINTS, BC_ZERO_FIRST, BC_ZERO_SECOND




from Peak1D cimport *
from Feature cimport *
from FeatureMap cimport *
from MSExperiment cimport *
from Peak1D cimport *
from ChromatogramPeak cimport *
from FeatureFinder cimport *
from DefaultParamHandler cimport *

cdef extern from "<OpenMS/TRANSFORMATIONS/FEATUREFINDER/FeatureFinderAlgorithmIsotopeWavelet.h>" namespace "OpenMS":

    cdef cppclass FeatureFinderAlgorithmIsotopeWavelet(DefaultParamHandler):

        # wrap-inherits:
        #    DefaultParamHandler
        FeatureFinderAlgorithmIsotopeWavelet()      nogil except +

        void setData(MSExperiment & input, FeatureMap& output, FeatureFinder & ff) nogil except +
        void run() nogil except +
        # MSSpectrum * createHRData(const UInt i)
        # static FeatureFinderAlgorithm* create()

cdef extern from "<OpenMS/TRANSFORMATIONS/FEATUREFINDER/FeatureFinderAlgorithmIsotopeWavelet.h>" namespace "OpenMS::FeatureFinderAlgorithmIsotopeWavelet":

    String getProductName()   nogil except + # wrap-attach:FeatureFinderAlgorithmIsotopeWavelet



from libcpp cimport bool
from libcpp.vector cimport vector as libcpp_vector
from Types cimport *
from DRange cimport *

cdef extern from "<OpenMS/FORMAT/OPTIONS/FeatureFileOptions.h>" namespace "OpenMS":

    cdef cppclass FeatureFileOptions:

        FeatureFileOptions() nogil except +
        FeatureFileOptions(FeatureFileOptions) nogil except +

        void setMetadataOnly(bool) nogil except +
        bool getMetadataOnly()     nogil except +

        void setSizeOnly(bool) nogil except +
        bool getSizeOnly()     nogil except +

        void setLoadConvexHull(bool) nogil except +
        bool getLoadConvexHull()     nogil except +

        void setLoadSubordinates(bool) nogil except +
        bool getLoadSubordinates()     nogil except +

        void setRTRange(DRange1 & range_) nogil except +
        bool hasRTRange() nogil except +
        DRange1 getRTRange() nogil except +
        void setMZRange(DRange1 & range_) nogil except +
        bool hasMZRange() nogil except +
        DRange1 getMZRange() nogil except +
        void setIntensityRange(DRange1 & range_) nogil except +
        bool hasIntensityRange() nogil except +
        DRange1 getIntensityRange() nogil except +




from Types cimport *
from TransformationModel cimport *
from BSpline2d cimport *

# ctypedef libcpp_vector[ libcpp_pair[double, double] ] DataPoints;

cdef extern from "<OpenMS/ANALYSIS/MAPMATCHING/TransformationModelBSpline.h>" namespace "OpenMS":

    cdef cppclass TransformationModelBSpline(TransformationModel) :
        # wrap-inherits:
        #  TransformationModel

        TransformationModelBSpline(TransformationModelBSpline) nogil except + #wrap-ignore

        TransformationModelBSpline(libcpp_vector[TM_DataPoint]& data, Param& params) nogil except +

        double evaluate(double value) nogil except +

cdef extern from "<OpenMS/ANALYSIS/MAPMATCHING/TransformationModelBSpline.h>" namespace "OpenMS::TransformationModelBSpline":

    void getDefaultParameters(Param& params) nogil except +  # wrap-attach:TransformationModelBSpline



from libcpp cimport bool
from Types cimport *
from DataValue cimport *
from Feature cimport *
from UniqueIdInterface cimport *
from MetaInfoInterface cimport *

cdef extern from "<OpenMS/METADATA/ProteinHit.h>" namespace "OpenMS":

    cdef cppclass ProteinHit(MetaInfoInterface):
        # wrap-inherits:
        #    MetaInfoInterface

        ProteinHit() nogil except +
        ProteinHit(double score, UInt rank, String accession, String sequence) nogil except +
        ProteinHit(ProteinHit) nogil except +

        # const members
        ## double COVERAGE_UNKNOWN

        float getScore() nogil except +
        UInt getRank() nogil except +
        String getSequence() nogil except +
        String getAccession() nogil except +
        String getDescription() nogil except +
        double getCoverage() nogil except +

        void setScore(float ) nogil except +
        void setRank(UInt) nogil except +
        void setSequence(String) nogil except +
        void setAccession(String) nogil except +
        void setDescription(String description) nogil except +
        void setCoverage(double) nogil except +

        bool operator==(ProteinHit) nogil except +
        bool operator!=(ProteinHit) nogil except +






from libcpp.vector cimport vector as libcpp_vector
from libcpp cimport bool
from ReactionMonitoringTransition cimport *
from TargetedExperimentHelper cimport *
from CVTermList cimport *
from IncludeExcludeTarget cimport *
from DataValue cimport *
from Software cimport *
from SourceFile cimport *

cdef extern from "<OpenMS/ANALYSIS/TARGETED/TargetedExperiment.h>" namespace "OpenMS":

    cdef cppclass TargetedExperiment:

        TargetedExperiment() nogil except +
        TargetedExperiment(TargetedExperiment &) nogil except +

        bool operator==(TargetedExperiment) nogil except +
        bool operator!=(TargetedExperiment) nogil except +

        TargetedExperiment operator+(TargetedExperiment)    nogil except +
        TargetedExperiment iadd(TargetedExperiment)   nogil except + # wrap-as:operator+=

        void clear(bool clear_meta_data)  nogil except +
        void sortTransitionsByProductMZ()  nogil except +

        # cv list
        void setCVs(libcpp_vector[CV] cvs) nogil except +
        libcpp_vector[CV] getCVs() nogil except +
        void addCV(CV cv) nogil except +

        # contact list
        void setContacts(libcpp_vector[Contact] contacts) nogil except +
        libcpp_vector[Contact] getContacts() nogil except +
        void addContact(Contact contact) nogil except +

        # publication list
        void setPublications(libcpp_vector[Publication] publications) nogil except +
        libcpp_vector[Publication] getPublications() nogil except +
        void addPublication(Publication publication) nogil except +

        # target list
        void setTargetCVTerms(CVTermList cv_terms) nogil except +
        CVTermList getTargetCVTerms() nogil except +
        void addTargetCVTerm(CVTerm cv_term) nogil except +
        void setTargetMetaValue(String name, DataValue value) nogil except +

        # instrument list
        void setInstruments(libcpp_vector[TargetedExperiment_Instrument] instruments) nogil except +
        libcpp_vector[TargetedExperiment_Instrument] getInstruments() nogil except +
        void addInstrument(TargetedExperiment_Instrument instrument) nogil except +

        # software list
        void setSoftware(libcpp_vector[Software] software) nogil except +
        libcpp_vector[Software] getSoftware() nogil except +
        void addSoftware(Software software) nogil except +

        # protein list
        void setProteins(libcpp_vector[Protein] proteins) nogil except +
        libcpp_vector[Protein] getProteins() nogil except +
        Protein getProteinByRef(String ref) nogil except +
        bool hasProtein(String ref) nogil except +
        void addProtein(Protein protein) nogil except +

        # compound list
        void setCompounds(libcpp_vector[Compound] rhs) nogil except +
        libcpp_vector[Compound] getCompounds() nogil except +
        void addCompound(Compound rhs) nogil except +
        bool hasCompound(String ref) nogil except +
        Compound getCompoundByRef(String ref) nogil except +

        # peptide list
        void setPeptides(libcpp_vector[Peptide] rhs) nogil except +
        libcpp_vector[Peptide] getPeptides() nogil except +
        bool hasPeptide(String ref) nogil except +
        Peptide getPeptideByRef(String ref) nogil except +
        void addPeptide(Peptide rhs) nogil except +

        # set transition list
        void setTransitions(libcpp_vector[ReactionMonitoringTransition] transitions) nogil except +
        libcpp_vector[ReactionMonitoringTransition] getTransitions() nogil except +
        void addTransition(ReactionMonitoringTransition transition) nogil except +

        void setIncludeTargets(libcpp_vector[IncludeExcludeTarget] targets) nogil except +
        libcpp_vector[IncludeExcludeTarget] getIncludeTargets() nogil except +
        void addIncludeTarget(IncludeExcludeTarget target) nogil except +
        void setExcludeTargets(libcpp_vector[IncludeExcludeTarget] targets) nogil except +
        libcpp_vector[IncludeExcludeTarget] getExcludeTargets() nogil except +
        void addExcludeTarget(IncludeExcludeTarget target) nogil except +

        # sets the source files
        void setSourceFiles(libcpp_vector[SourceFile] source_files) nogil except +
        libcpp_vector[SourceFile] getSourceFiles() nogil except +
        void addSourceFile(SourceFile source_file) nogil except +

        bool containsInvalidReferences() nogil except +




from Types cimport *
from ConsensusIDAlgorithmSimilarity cimport *

cdef extern from "<OpenMS/ANALYSIS/ID/ConsensusIDAlgorithmPEPMatrix.h>" namespace "OpenMS":
    
    cdef cppclass ConsensusIDAlgorithmPEPMatrix(ConsensusIDAlgorithmSimilarity) :
        # wrap-inherits:
        #  ConsensusIDAlgorithmSimilarity
        ConsensusIDAlgorithmPEPMatrix() nogil except +
        ConsensusIDAlgorithmPEPMatrix(ConsensusIDAlgorithmPEPMatrix) nogil except + #wrap-ignore




from Types cimport *
from libcpp cimport bool
from MzMLSqliteHandler cimport *
from Types cimport *
from MSExperiment cimport *
from IMSDataConsumer cimport *

cdef extern from "<OpenMS/FORMAT/SqMassFile.h>" namespace "OpenMS":
    
    cdef cppclass SqMassFile:
        SqMassFile() nogil except +
        SqMassFile(SqMassFile) nogil except + #wrap-ignore
        void load(const String & filename, MSExperiment & map_) nogil except +
        void store(const String & filename, MSExperiment & map_) nogil except +
        # NAMESPACE # # POINTER # void transform(const String & filename_in, Interfaces::IMSDataConsumer * consumer, bool skip_full_count, bool skip_first_pass) nogil except +
        void setConfig(SqMassConfig config) nogil except +

cdef extern from "<OpenMS/FORMAT/SqMassFile.h>" namespace "OpenMS::SqMassFile":
    
    cdef cppclass SqMassConfig "OpenMS::SqMassFile::SqMassConfig":
        SqMassConfig() nogil except +
        SqMassConfig(SqMassConfig) nogil except + #wrap-ignore
        bool write_full_meta
        bool use_lossy_numpress
        double linear_fp_mass_acc




from Types cimport *
from libcpp cimport bool
from libcpp.vector cimport vector as libcpp_vector
from DefaultParamHandler cimport *
from MSSpectrum cimport *
from AASequence cimport *
from ProteinProteinCrossLink cimport *

cdef extern from "<OpenMS/CHEMISTRY/SimpleTSGXLMS.h>" namespace "OpenMS":

    cdef cppclass SimpleTSGXLMS(DefaultParamHandler) :
        # wrap-inherits:
        #  DefaultParamHandler

        SimpleTSGXLMS() nogil except +
        SimpleTSGXLMS(SimpleTSGXLMS) nogil except +

        void getLinearIonSpectrum(libcpp_vector[ SimplePeak ]& spectrum, AASequence peptide,
                Size link_pos, int charge, Size link_pos_2) nogil except +

        void getXLinkIonSpectrum(libcpp_vector[ SimplePeak ]& spectrum, AASequence peptide,
                Size link_pos, double precursor_mass,
                int mincharge, int maxcharge, Size link_pos_2) nogil except +

        void getXLinkIonSpectrum(libcpp_vector[ SimplePeak ]& spectrum, ProteinProteinCrossLink crosslink,
                bool frag_alpha, int mincharge, int maxcharge) nogil except +


cdef extern from "<OpenMS/ANALYSIS/XLMS/OPXLDataStructs.h>" namespace "OpenMS::SimpleTSGXLMS":

    cdef cppclass SimplePeak "OpenMS::SimpleTSGXLMS::SimplePeak":

        SimplePeak(SimplePeak) nogil except +
        SimplePeak() nogil except +
        SimplePeak(double mz, int charge) nogil except +

        double mz
        int charge



from String cimport *
from Types cimport *

cdef extern from "<OpenMS/CONCEPT/ProgressLogger.h>" namespace "OpenMS":

    cdef cppclass ProgressLogger:
        ProgressLogger()           nogil except +
        void setLogType(LogType)           nogil except +
        LogType getLogType()           nogil except +
        void startProgress(SignedSize begin, SignedSize end, String label)           nogil except +
        void setProgress(SignedSize value)           nogil except +
        void endProgress() nogil except +
        void nextProgress() nogil except +


cdef extern from "<OpenMS/CONCEPT/ProgressLogger.h>" namespace "OpenMS::ProgressLogger":

    cdef enum LogType:
        # wrap-attach: ProgressLogger
        CMD, GUI, NONE





from Fitter1D cimport *

cdef extern from "<OpenMS/TRANSFORMATIONS/FEATUREFINDER/LevMarqFitter1D.h>" namespace "OpenMS":
    
    cdef cppclass LevMarqFitter1D(Fitter1D):
        # wrap-ignore
        # no-pxd-import
        LevMarqFitter1D() nogil except +
        LevMarqFitter1D(LevMarqFitter1D) nogil except +




from Types cimport *
from libcpp.map cimport map as libcpp_map
from libcpp.vector cimport vector as libcpp_vector
from DefaultParamHandler cimport *
from String cimport *

ctypedef libcpp_map[String, libcpp_vector[double] ] PredictorMap

cdef extern from "<OpenMS/ANALYSIS/SVM/SimpleSVM.h>" namespace "OpenMS":
    
    cdef cppclass SimpleSVM(DefaultParamHandler) :
        # wrap-inherits:
        #  DefaultParamHandler
        SimpleSVM() nogil except +
        SimpleSVM(SimpleSVM) nogil except + #wrap-ignore

        # Currently nested things inside std::map doesnt work
        # void setup(libcpp_map[String, libcpp_vector[double] ]& predictors, libcpp_map[ Size, Int ] & labels) nogil except +
        void predict(libcpp_vector[ SVMPrediction ] & predictions, libcpp_vector[ size_t ] indexes) nogil except +
        void getFeatureWeights(libcpp_map[ String, double ] & feature_weights) nogil except +
        void writeXvalResults(const String & path) nogil except +


cdef extern from "<OpenMS/ANALYSIS/SVM/SimpleSVM.h>" namespace "OpenMS::SimpleSVM":
    
    cdef cppclass SVMPrediction "OpenMS::SimpleSVM::Prediction":

        SVMPrediction() nogil except +
        SVMPrediction(SVMPrediction) nogil except + #wrap-ignore

        Int label
        libcpp_map[int, double ] probabilities




from Types cimport *
from FeatureMap cimport *
from ConsensusMap cimport *
from ILPDCWrapper cimport *
from DPosition cimport *
from MassExplainer cimport *
from DefaultParamHandler cimport *

cdef extern from "<OpenMS/ANALYSIS/DECHARGING/MetaboliteFeatureDeconvolution.h>" namespace "OpenMS":
    
    cdef cppclass MetaboliteFeatureDeconvolution(DefaultParamHandler) :
        # wrap-inherits:
        #  DefaultParamHandler
        MetaboliteFeatureDeconvolution() nogil except +
        MetaboliteFeatureDeconvolution(MetaboliteFeatureDeconvolution) nogil except +
        void compute(FeatureMap & fm_in, FeatureMap & fm_out, ConsensusMap & cons_map, ConsensusMap & cons_map_p) nogil except +

cdef extern from "<OpenMS/ANALYSIS/DECHARGING/MetaboliteFeatureDeconvolution.h>" namespace "OpenMS::MetaboliteFeatureDeconvolution":

    cdef enum CHARGEMODE_MFD "OpenMS::MetaboliteFeatureDeconvolution::CHARGEMODE":
        #wrap-attach:
        #    MetaboliteFeatureDeconvolution
        #wrap-instances:
        #    CHARGEMODE := CHARGEMODE_MFD

        QFROMFEATURE
        QHEURISTIC
        QALL




from Types  cimport *
from smart_ptr cimport shared_ptr
from libcpp.string cimport string as libcpp_string
from libcpp.vector cimport vector as libcpp_vector
from OpenSwathDataStructures cimport *

# ctypedef shared_ptr[ISpectrumAccess] SpectrumAccessPtr
cdef extern from "<OpenMS/OPENSWATHALGO/DATAACCESS/ISpectrumAccess.h>" namespace "OpenSwath":

  # This is an abstract base class in C++ with only pure virtual functions -> we can thus not create an instance of it
  cdef cppclass ISpectrumAccess:
      # wrap-ignore
      # ABSTRACT class
      # no-pxd-import

      ISpectrumAccess() nogil except +
      ISpectrumAccess(ISpectrumAccess) nogil except +

      # virtual boost::shared_ptr<ISpectrumAccess> lightClone() const = 0;

      shared_ptr[OSSpectrum] getSpectrumById(int id_) nogil except +
      libcpp_vector[size_t] getSpectraByRT(double RT, double deltaRT) nogil except +
      size_t getNrSpectra() nogil except +
      # virtual SpectrumMeta getSpectrumMetaById(int id) const = 0;

      shared_ptr[OSChromatogram] getChromatogramById(int id_) nogil except +
      size_t getNrChromatograms() nogil except +
      libcpp_string getChromatogramNativeID(int id_) nogil except +




from MSExperiment cimport *
from MSChromatogram cimport *
from Peak1D cimport *
from ChromatogramPeak cimport *

from MassTrace cimport *
from Feature cimport *
from FeatureMap cimport *
from MSChromatogram cimport *

from DefaultParamHandler cimport *
from ProgressLogger cimport *

cdef extern from "<OpenMS/FILTERING/DATAREDUCTION/FeatureFindingMetabo.h>" namespace "OpenMS":

    cdef cppclass FeatureFindingMetabo(ProgressLogger, DefaultParamHandler):
        # wrap-inherits:
        #    ProgressLogger
        #    DefaultParamHandler
        #

        FeatureFindingMetabo()      nogil except +

        void run(libcpp_vector[Kernel_MassTrace] input_mtraces,
                 FeatureMap & output_featmap,
                 libcpp_vector[ libcpp_vector[ MSChromatogram ] ] & output_chromatograms
                 ) nogil except +




from Types cimport *
from ProteinIdentification cimport *
from PeptideIdentification cimport *

cdef extern from "<OpenMS/FORMAT/OMSSACSVFile.h>" namespace "OpenMS":
    
    cdef cppclass OMSSACSVFile "OpenMS::OMSSACSVFile":
        OMSSACSVFile() nogil except +
        OMSSACSVFile(OMSSACSVFile) nogil except + #wrap-ignore

        void load(const String & filename,
                  ProteinIdentification & protein_identification,
                  libcpp_vector[ PeptideIdentification ] & id_data) nogil except +





from OPXLDataStructs cimport *
from String cimport String
from libcpp cimport bool
from ResidueModification cimport TermSpecificity
from AASequence cimport AASequence


cdef extern from "<OpenMS/ANALYSIS/XLMS/OPXLDataStructs.h>" namespace "OpenMS::OPXLDataStructs":

    cdef cppclass ProteinProteinCrossLink "OpenMS::OPXLDataStructs::ProteinProteinCrossLink":

        ProteinProteinCrossLink(ProteinProteinCrossLink) nogil except +
        ProteinProteinCrossLink() nogil except +


        const AASequence *alpha
        const AASequence *beta
        libcpp_pair[ ptrdiff_t, ptrdiff_t] cross_link_position
        double cross_linker_mass
        String cross_linker_name
        TermSpecificity term_spec_alpha
        TermSpecificity term_spec_beta
        int precursor_correction

        ProteinProteinCrossLinkType getType() nogil except +
        bool operator==(ProteinProteinCrossLink other) nogil except +



import copy, sys
import pyopenms

"""
python pyTOPP/MRMMapper.py --in ../source/TEST/TOPP/MRMMapping_input.chrom.mzML  \
        --tr ../source/TEST/TOPP/MRMMapping_input.TraML --out MRMMapper.tmp.out \
        && diff MRMMapper.tmp.out ../source/TEST/TOPP/MRMMapping_output.chrom.mzML

"""

def algorithm(chromatogram_map, targeted, precursor_tolerance, product_tolerance, allow_unmapped=True, allow_double_mappings=False):

    output = copy.copy(chromatogram_map)
    output.clear(False);
    empty_chromats = []
    output.setChromatograms(empty_chromats);

    notmapped = 0
    for chrom in chromatogram_map.getChromatograms():
        mapped_already = False
        for transition in targeted.getTransitions():
            if (abs(chrom.getPrecursor().getMZ() - transition.getPrecursorMZ()) < precursor_tolerance and
                abs(chrom.getProduct().getMZ()  -  transition.getProductMZ()) < product_tolerance):
                if mapped_already:
                    this_peptide = targeted.getPeptideByRef(transition.getPeptideRef() ).sequence
                    other_peptide = chrom.getPrecursor().getMetaValue("peptide_sequence")
                    print "Found mapping of", chrom.getPrecursor().getMZ(), "/", chrom.getProduct().getMZ(), "to", transition.getPrecursorMZ(), "/",transition.getProductMZ()
                    print "Of peptide", this_peptide
                    print "But the chromatogram is already mapped to", other_peptide
                    if not allow_double_mappings: raise Exception("Cannot map twice")
                mapped_already = True
                precursor = chrom.getPrecursor();
                peptide = targeted.getPeptideByRef(transition.getPeptideRef() )
                precursor.setMetaValue("peptide_sequence", peptide.sequence)
                chrom.setPrecursor(precursor)
                chrom.setNativeID(transition.getNativeID())
        if not mapped_already:
            notmapped += 1
            print "Did not find a mapping for chromatogram", chrom.getNativeID()
            if not allow_unmapped: raise Exception("No mapping")
        else:
            output.addChromatogram(chrom)

    if notmapped > 0:
        print "Could not find mapping for", notmapped, "chromatogram(s)"


    dp = pyopenms.DataProcessing()
    # dp.setProcessingActions(ProcessingAction:::FORMAT_CONVERSION)
    pa = pyopenms.ProcessingAction().FORMAT_CONVERSION
    dp.setProcessingActions(set([pa]))

    chromatograms = output.getChromatograms();
    for chrom in chromatograms:
        this_dp = chrom.getDataProcessing()
        this_dp.append(dp)
        chrom.setDataProcessing(this_dp)

    output.setChromatograms(chromatograms);
    return output

def main(options):
    precursor_tolerance = options.precursor_tolerance
    product_tolerance = options.product_tolerance
    out = options.outfile
    chromat_in = options.infile
    traml_in = options.traml_in

    # precursor_tolerance = 0.05
    # product_tolerance = 0.05
    # out = "/tmp/out.mzML"
    # chromat_in = "../source/TEST/TOPP/MRMMapping_input.chrom.mzML"
    # traml_in = "../source/TEST/TOPP/MRMMapping_input.TraML"

    ff = pyopenms.MRMFeatureFinderScoring()
    chromatogram_map = pyopenms.MSExperiment()
    fh = pyopenms.FileHandler()
    fh.loadExperiment(chromat_in, chromatogram_map)
    targeted = pyopenms.TargetedExperiment();
    tramlfile = pyopenms.TraMLFile();
    tramlfile.load(traml_in, targeted);

    output = algorithm(chromatogram_map, targeted, precursor_tolerance, product_tolerance)

    pyopenms.MzMLFile().store(out, output);

def handle_args():
    import argparse

    usage = ""
    usage += "\nMRMMapper maps measured chromatograms (mzML) and the transitions used (TraML)"

    parser = argparse.ArgumentParser(description = usage )
    parser.add_argument('--in', dest="infile", help = 'An input file containing chromatograms')
    parser.add_argument("--tr", dest="traml_in", help="TraML input file containt the transitions")
    parser.add_argument("--out", dest="outfile", help="Output file with annotated chromatograms")
    parser.add_argument("--precursor_tolerance", dest="precursor_tolerance", default=0.1, help="Precursor tolerance when mapping (in Th)", metavar='0.1', type=float)
    parser.add_argument("--product_tolerance", dest="product_tolerance", default=0.1, help="Product tolerance when mapping (in Th)", metavar='0.1', type=float)

    args = parser.parse_args(sys.argv[1:])
    return args

if __name__ == '__main__':
    options = handle_args()
    main(options)



from libcpp.vector cimport vector as libcpp_vector
from libcpp.set cimport set as libcpp_set
from libcpp cimport bool

from String cimport *
from DefaultParamHandler cimport *

from PeptideIdentification cimport *
from ProteinIdentification cimport *

cdef extern from "<OpenMS/FORMAT/FASTAFile.h>" namespace "OpenMS":

    cdef cppclass FASTAFile:

        FASTAFile() nogil except +
        FASTAFile(FASTAFile) nogil except + #wrap-ignore

        void load(const String& filename, libcpp_vector[FASTAEntry] & data) nogil except +
        void store(const String& filename, libcpp_vector[FASTAEntry] & data) nogil except +

        void readStart(const String & filename) nogil except +
        bool readNext(FASTAEntry & protein) nogil except +
        # NAMESPACE # std::streampos position() nogil except +
        bool atEnd() nogil except +
        # NAMESPACE # bool setPosition(const std::streampos & pos) nogil except +
        void writeStart(const String & filename) nogil except +
        void writeNext(const FASTAEntry & protein) nogil except +
        void writeEnd() nogil except +

cdef extern from "<OpenMS/FORMAT/FASTAFile.h>" namespace "OpenMS::FASTAFile":

    cdef cppclass FASTAEntry:
        FASTAEntry() nogil except +
        FASTAEntry(FASTAEntry) nogil except +

        String identifier
        String description
        String sequence

        bool headerMatches(const FASTAEntry & rhs) nogil except +
        bool sequenceMatches(const FASTAEntry & rhs) nogil except +



import argparse
import pyopenms as pms
import logging
from common import addDataProcessing, writeParamsIfRequested, updateDefaults


def run_peak_picker(input_map, params, out_path):

    spec0 = input_map[0]
    if pms.PeakTypeEstimator().estimateType(spec0) == \
        pms.                               SpectrumSettings.SpectrumType.PEAKS:
        logging.warn("input peak map does not look like profile data")

    if any(not s.isSorted() for s in input_map):
        raise Exception("Not all spectra are sorted according to m/z")

    pp = pms.PeakPickerHiRes()
    pp.setParameters(params)
    out_map = pms.MSExperiment()
    pp.pickExperiment(input_map, out_map)

    out_map = addDataProcessing(out_map, params, pms.ProcessingAction.PEAK_PICKING)
    fh = pms.FileHandler()
    fh.storeExperiment(out_path, out_map)


def main():

    parser = argparse.ArgumentParser(description="PeakPickerHiRes")
    parser.add_argument("-in",
                        action="store",
                        type=str,
                        dest="in_",
                        metavar="input_file",
                        )

    parser.add_argument("-out",
                        action="store",
                        type=str,
                        metavar="output_file",
                        )

    parser.add_argument("-ini",
                        action="store",
                        type=str,
                        metavar="ini_file",
                        )

    parser.add_argument("-dict_ini",
                        action="store",
                        type=str,
                        metavar="python_dict_ini_file",
                        )

    parser.add_argument("-write_ini",
                        action="store",
                        type=str,
                        metavar="ini_file",
                        )

    parser.add_argument("-write_dict_ini",
                        action="store",
                        type=str,
                        metavar="python_dict_ini_file",
                        )

    args = parser.parse_args()

    run_mode = args.in_ is not None and args.out is not None\
                and (args.ini is not None or args.dict_ini is not None)
    write_mode = args.write_ini is not None or args.write_dict_ini is not None
    ok = run_mode or write_mode
    if not ok:
        parser.error("either specify -in, -out and -(dict)ini for running "
                     "the peakpicker\nor -write(dict)ini for creating std "
                     "ini file")

    defaults = pms.PeakPickerHiRes().getDefaults()

    write_requested = writeParamsIfRequested(args, defaults)

    if not write_requested:
        updateDefaults(args, defaults)

        fh = pms.MzMLFile()
        fh.setLogType(pms.LogType.CMD)
        input_map = pms.MSExperiment()
        fh.load(args.in_, input_map)

        run_peak_picker(input_map, defaults, args.out)

if __name__ == "__main__":
    main()



from Types cimport *
from libcpp.vector cimport vector as libcpp_vector
from OpenSwathDataStructures cimport *
from ISpectrumAccess cimport *
from MSExperiment cimport *

cdef extern from "<OpenMS/ANALYSIS/OPENSWATH/DATAACCESS/SpectrumAccessOpenMS.h>" namespace "OpenMS":

  cdef cppclass SpectrumAccessOpenMS(ISpectrumAccess):
        # wrap-inherits:
        #  ISpectrumAccess

        SpectrumAccessOpenMS() # wrap-pass-constructor

        SpectrumAccessOpenMS(SpectrumAccessOpenMS)
        SpectrumAccessOpenMS(shared_ptr[ MSExperiment ] & ms_experiment) nogil except +




from Types cimport *
from ProgressLogger cimport *
from DefaultParamHandler cimport *
from MSSpectrum cimport *
from Peak1D cimport *
from MSExperiment cimport *
from MzTab cimport *
from StringList cimport *
from libcpp cimport bool

cdef extern from "<OpenMS/ANALYSIS/RNPXL/RNPxlModificationsGenerator.h>" namespace "OpenMS":

    cdef cppclass RNPxlModificationsGenerator:

        RNPxlModificationsGenerator() nogil except +
        RNPxlModificationsGenerator(RNPxlModificationsGenerator) nogil except + 

        RNPxlModificationMassesResult initModificationMassesRNA(
            StringList target_nucleotides,
            StringList nt_groups,
            libcpp_set[char] can_xl,
            StringList mappings,
            StringList modifications,
            String sequence_restriction, 
            bool cysteine_adduct, 
            Int max_length) nogil except +

cdef extern from "<OpenMS/ANALYSIS/RNPXL/RNPxlModificationsGenerator.h>" namespace "OpenMS":

    cdef cppclass RNPxlModificationMassesResult:

        RNPxlModificationMassesResult() nogil except +
        RNPxlModificationMassesResult(RNPxlModificationMassesResult) nogil except + 

        # libcpp_map[String, double] mod_masses # empirical formula -> mass
        # libcpp_map[String, libcpp_set[String] ] mod_combinations # empirical formula -> nucleotide formula(s) (formulas if modifications lead to ambiguities)
        # libcpp_map[Size, String] mod_formula_idx





from libcpp.map cimport map as libcpp_map
from libcpp.pair cimport pair as libcpp_pair
from libcpp cimport bool
from Types cimport *
from ProgressLogger cimport *
from DefaultParamHandler cimport *
from FASTAFile cimport *
from MetaInfoInterface cimport *

cdef extern from "<OpenMS/SIMULATION/SimTypes.h>" namespace "OpenMS::SimTypes":

    cdef cppclass SimRandomNumberGenerator:

        SimRandomNumberGenerator() nogil except +
        SimRandomNumberGeneratorSimTypes(SimRandomNumberGeneratorSimTypes) nogil except + # wrap-ignore
        void initialize(bool biological_random, bool technical_random) nogil except +

    cdef cppclass SimProtein:
        SimProtein(FASTAEntry entry, MetaInfoInterface meta)
        # TODO does this work? 
        # FASTAFile::FASTAEntry entry
        # MetaInfoInterface meta

    ctypedef libcpp_vector[SimProtein] SampleProteins
    ctypedef libcpp_vector[libcpp_vector[SimProtein]] SampleChannels



from Types cimport *
from FeatureFinderAlgorithmPickedHelperStructs cimport *

cdef extern from "<OpenMS/FILTERING/DATAREDUCTION/IsotopeDistributionCache.h>" namespace "OpenMS":
    
    cdef cppclass IsotopeDistributionCache "OpenMS::IsotopeDistributionCache":
        IsotopeDistributionCache(IsotopeDistributionCache) nogil except + #wrap-ignore
        IsotopeDistributionCache(double max_mass, double mass_window_width, double intensity_percentage, double intensity_percentage_optional) nogil except +
        TheoreticalIsotopePattern  getIsotopeDistribution(double mass) nogil except +




from libcpp cimport bool
from libcpp.vector cimport vector as libcpp_vector
from libcpp.set cimport set as libcpp_set
from DataValue cimport *
from String cimport *
from Types cimport *
from MetaInfoInterface cimport *
from PeptideHit cimport *

cdef extern from "<OpenMS/METADATA/PeptideIdentification.h>" namespace "OpenMS":

    cdef cppclass PeptideIdentification(MetaInfoInterface):
        # wrap-inherits:
        #    MetaInfoInterface

        PeptideIdentification() nogil except +
        PeptideIdentification(PeptideIdentification) nogil except +
        bool operator==(PeptideIdentification) nogil except +
        bool operator!=(PeptideIdentification) nogil except +

        libcpp_vector[PeptideHit] getHits() nogil except +
        void insertHit(PeptideHit) nogil except +
        void setHits(libcpp_vector[PeptideHit]) nogil except +

        double getSignificanceThreshold()   nogil except +
        # setting of the peptide significance threshold value
        void setSignificanceThreshold(double value) nogil except +

        String     getScoreType() nogil except +
        void       setScoreType(String) nogil except +
        bool       isHigherScoreBetter() nogil except +
        void       setHigherScoreBetter(bool) nogil except +
        String     getIdentifier() nogil except +
        void       setIdentifier(String) nogil except +

        bool       hasMZ() nogil except +
        double     getMZ() nogil except +
        void       setMZ(double) nogil except +

        bool       hasRT() nogil except +
        double     getRT() nogil except +
        void       setRT(double) nogil except +

        String     getBaseName() nogil except +
        void       setBaseName(String) nogil except +

        String     getExperimentLabel() nogil except +
        void       setExperimentLabel(String) nogil except +

        void       assignRanks() nogil except +
        void       sort() nogil except +
        void sortByRank() nogil except +
        bool       empty() nogil except +

        libcpp_vector[PeptideHit] getReferencingHits(libcpp_vector[PeptideHit], libcpp_set[String] &) nogil except +




from Types cimport *
from libcpp cimport bool
from IMSIsotopeDistribution cimport *
from libcpp.string cimport string as libcpp_string

cdef extern from "<OpenMS/CHEMISTRY/MASSDECOMPOSITION/IMS/IMSElement.h>" namespace "OpenMS::ims::IMSElement":

    ctypedef libcpp_string name_type
    ctypedef IMSIsotopeDistribution isotopes_type

cdef extern from "<OpenMS/CHEMISTRY/MASSDECOMPOSITION/IMS/IMSElement.h>" namespace "OpenMS::ims":
    
    cdef cppclass IMSElement "OpenMS::ims::IMSElement":
        IMSElement() nogil except +
        IMSElement(IMSElement) nogil except +
        # mass_type ELECTRON_MASS_IN_U read-only
        IMSElement(name_type & name, isotopes_type & isotopes) nogil except +
        IMSElement(name_type & name, mass_type mass) nogil except +
        IMSElement(name_type & name, nominal_mass_type nominal_mass) nogil except +
        name_type  getName() nogil except +
        void setName(name_type & name) nogil except +
        name_type  getSequence() nogil except +
        void setSequence(name_type & sequence) nogil except +
        nominal_mass_type getNominalMass() nogil except +
        mass_type getMass(size_type index) nogil except +
        mass_type getAverageMass() nogil except +
        mass_type getIonMass(int electrons_number) nogil except +
        IMSIsotopeDistribution  getIsotopeDistribution() nogil except +
        void setIsotopeDistribution(IMSIsotopeDistribution & isotopes) nogil except +
        bool operator==(IMSElement & element) nogil except +
        bool operator!=(IMSElement & element) nogil except +




from Types cimport *
from KDTreeFeatureMaps cimport *
from TransformationModelLowess cimport *

cdef extern from "<OpenMS/ANALYSIS/MAPMATCHING/MapAlignmentAlgorithmKD.h>" namespace "OpenMS":
    
    cdef cppclass MapAlignmentAlgorithmKD "OpenMS::MapAlignmentAlgorithmKD":
        MapAlignmentAlgorithmKD(MapAlignmentAlgorithmKD) nogil except + #wrap-ignore
        MapAlignmentAlgorithmKD(Size num_maps, Param & param) nogil except +
        void addRTFitData(KDTreeFeatureMaps & kd_data) nogil except +
        void fitLOWESS() nogil except +
        void transform(KDTreeFeatureMaps & kd_data) nogil except +




from Types cimport *
from MassTrace cimport *
from Feature cimport *
from ConsensusFeature cimport *
from FeatureMap cimport *
from ConsensusMap cimport *
from MzTab cimport *
from DefaultParamHandler cimport *
from String cimport *
from EmpiricalFormula cimport *

cdef extern from "<OpenMS/ANALYSIS/ID/AccurateMassSearchEngine.h>" namespace "OpenMS":
    
    cdef cppclass AMSE_AdductInfo "OpenMS::AdductInfo":

        AMSE_AdductInfo(AMSE_AdductInfo) nogil except + #wrap-ignore
        AMSE_AdductInfo(const String & name, EmpiricalFormula & adduct, int charge, UInt mol_multiplier) nogil except +

        double getNeutralMass(double observed_mz) nogil except +
        double getMZ(double neutral_mass) nogil except +
        bool isCompatible(EmpiricalFormula db_entry) nogil except +
        int getCharge() nogil except +
        String getName() nogil except +

        # TODO: AMSE_AdductInfo has private constructor
        # AMSE_AdductInfo parseAdductString(const String & adduct) nogil except +




from SourceFile cimport *
from DateTime cimport *
from DocumentIdentifier cimport *
from MetaInfoInterface cimport *
from libcpp.vector cimport vector as libcpp_vector

cdef extern from "<OpenMS/METADATA/Sample.h>" namespace "OpenMS":

    cdef cppclass Sample(MetaInfoInterface):
        # wrap-inherits:
        #    MetaInfoInterface

        Sample() nogil except +
        Sample(Sample) nogil except + # wrap-ignore

        #returns the sample name (default: "")
        String getName() nogil except +
        #sets the sample name
        void setName(String name) nogil except +

        #returns the sample name (default: "")
        String getOrganism() nogil except +
        #sets the sample name
        void setOrganism(String organism) nogil except +

        # returns the sample number (default: "")
        String getNumber() nogil except +
        # sets the sample number (e.g. sample ID)
        void setNumber(String number) nogil except +

        # returns the comment (default: "")
        String getComment() nogil except +
        # sets the comment (may contain newline characters)
        void setComment(String comment) nogil except +

        # returns the state of aggregation (default: SAMPLENULL)
        SampleState getState() nogil except +
        # sets the state of aggregation
        void setState(SampleState state) nogil except +

        # returns the mass (in gram) (default: 0.0)
        double getMass() nogil except +
        # sets the mass (in gram)
        void setMass(double mass) nogil except +

        # returns the volume (in ml) (default: 0.0)
        double getVolume() nogil except +
        # sets the volume (in ml)
        void setVolume(double volume) nogil except +

        # returns the concentration (in g/l) (default: 0.0)
        double getConcentration() nogil except +
        # sets the concentration (in g/l)
        void setConcentration(double concentration) nogil except +

        # returns a reference to the vector of subsamples that were combined to create this sample
        libcpp_vector[Sample] getSubsamples() nogil except +
        # sets the vector of subsamples that were combined to create this sample
        void setSubsamples(libcpp_vector[Sample] subsamples) nogil except +

        # Since SampleTreatment is abstract, we cant wrap it
        ## void addTreatment(SampleTreatment treatment, Int before_position = -1) nogil except +
        ## SampleTreatment getTreatment(UInt position) nogil except +

        # brief removes the sample treatment at the given position
        void removeTreatment(UInt position) nogil except +
        # returns the number of sample treatments
        Int countTreatments() nogil except +

cdef extern from "<OpenMS/METADATA/Sample.h>" namespace "OpenMS::Sample":

    cdef enum SampleState:
        # wrap-attach:
        #     Sample
    
        SAMPLENULL, SOLID, LIQUID, GAS, SOLUTION, EMULSION, SUSPENSION, SIZE_OF_SAMPLESTATE




from .special_autowrap_conversionproviders import *
from autowrap.ConversionProvider import special_converters

def register_converters():

    special_converters.append(OpenMSStringConverter())
    special_converters.append(StdVectorStringConverter())
    special_converters.append(StdSetStringConverter())
    special_converters.append(OpenMSIntListConverter())
    special_converters.append(OpenMSStringListConverter())
    special_converters.append(OpenMSDoubleListConverter())
    special_converters.append(OpenMSMapConverter())
    special_converters.append(CVTermMapConverter())
    special_converters.append(OpenMSDataValue())
    special_converters.append(OpenMSDPosition2())
    special_converters.append(OpenMSDPosition2Vector())



from MSSpectrum cimport *
from MSChromatogram cimport *
from ExperimentalSettings cimport *

cdef extern from "<OpenMS/FORMAT/DATAACCESS/MSDataCachedConsumer.h>" namespace "OpenMS":

    cdef cppclass MSDataCachedConsumer:

        MSDataCachedConsumer(String filename) nogil except +
        MSDataCachedConsumer(String filename, bool clear) nogil except +
        MSDataCachedConsumer(MSDataCachedConsumer) nogil except + #wrap-ignore

        void consumeSpectrum(MSSpectrum & s) nogil except +
        void consumeChromatogram(MSChromatogram & c) nogil except +

        void setExperimentalSettings(ExperimentalSettings& exp) nogil except +
        void setExpectedSize(Size expectedSpectra, Size expectedChromatograms) nogil except +




#!/usr/bin/env python
# -*- coding: utf-8  -*-
"""
--------------------------------------------------------------------------
                  OpenMS -- Open-Source Mass Spectrometry
--------------------------------------------------------------------------
Copyright The OpenMS Team -- Eberhard Karls University Tuebingen,
ETH Zurich, and Freie Universitaet Berlin 2002-2020.

This software is released under a three-clause BSD license:
 * Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
 * Neither the name of any author or any participating institution
   may be used to endorse or promote products derived from this software
   without specific prior written permission.
For a full list of authors, refer to the file AUTHORS.
--------------------------------------------------------------------------
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL ANY OF THE AUTHORS OR THE CONTRIBUTING
INSTITUTIONS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
--------------------------------------------------------------------------
$Maintainer: Hannes Roest$
$Authors: Hannes Roest$
--------------------------------------------------------------------------
"""
from __future__ import print_function

import glob, os, sys
import re, time
import argparse
from xml.sax.saxutils import escape as xml_escape

from PythonCheckerLib import parse_pxd_file
from PythonCheckerLib import create_pxd_file_map

# Try non-standard libs
try:
    import yaml
    import breathe
    from Cython.Compiler.Nodes import CEnumDefNode, CppClassNode, CTypeDefNode, CVarDefNode, CImportStatNode, CDefExternNode
    from autowrap.PXDParser import CppClassDecl, CTypeDefDecl, MethodOrAttributeDecl, EnumDecl
except ImportError as e:
    print ("You need to install a few packages for this library to work")
    print ("Please use:")
    print (" pip install breathe")
    print (" pip install pyyaml")
    print (" pip install autowrap")
    print (" pip install Cython")
    raise e

# Try breathe parser
try:
    from breathe.parser.eoxygen.compound import parse as doxygen_parse
except ImportError:
    print ("importing breathe.parser.doxygen.compound failed, try new API")
    from breathe.parser.compound import parse as doxygen_parse


# Matching to match doxygen methods to Cython pxd functions
def handle_member_definition(mdef, pxd_class, cnt):
    """ Matches a doxygen member definition (mdef) to a Cython pxd file.

    This tries to ensure that all C++ functions are wrapped and have an
    equivalent in the Python wrapper.

    Parameters
    ----------
    mdef : breathe.parser.compound.memberdefTypeSub
        A doxygen entry
    pxd_class : autowrap.PXDParser.CppClassDecl
        A PXD class file as parsed by autowrap
    cnt :
        A count object to keep track of how many functions we wrapped
    """

    tres = TestResult()
    protection = mdef.get_prot() # DoxProtectionKind: public, protected, private, package
    kind = mdef.get_kind() # DoxMemberKind: define property event variable typedef enum function signal prototype friend dcop slot

    if not protection in "public protected private package".split(" "):
        raise Exception("Error; something is wrong")

    if not kind in "variable enum function define property event typedef signal prototype friend dcop slot".split(" "):
        raise Exception("Error; something is wrong")

    # Only match public enums, variables, functions
    if protection in "protected private package".split(" "):
        tres.setPassed(True)
    elif kind in "define property event typedef signal prototype friend dcop slot".split(" "):
        tres.setPassed(True)
    elif kind == "enum" and protection == "public":
        cnt.public_enums_total += 1
        cython_file = parse_pxd_file(pxd_class.pxd_path)
        found = False
        for klass in cython_file:

            if hasattr(klass[0], "name") and klass[0].name == mdef.get_name():
                found = True
                break

            # Sometimes we rename things in pyOpenMS for sanity (and namespace consistency) sake
            # E.g. OpenMS::PercolatorOutfile::ScoreType becomes PercolatorOutfile_ScoreType
            # and we have to go back to the full cname. However, the doxygen name needs to be inferred
            if hasattr(klass[0], "cname") and klass[0].cname.endswith(mdef.get_name()):
                assumed_fullname = mdef.compoundname + "::" + mdef.get_name()
                if (assumed_fullname == klass[0].cname):
                    found = True
                    break
                else:
                    print ("Something went wrong, %s is not equal to %s" % (assumed_fullname, klass[0].cname))

        if not found:
            tres.setPassed(False)
            tres.setMessage("TODO: Found enum in C++ but not in pxd: %s %s %s" % (mdef.kind, mdef.prot, mdef.name))
            cnt.public_enums_missing += 1
            comp_name = mdef.parent_doxy_file.compound.get_compoundname()
            internal_file_name = mdef.parent_doxy_file.getInternalFileName()
            namespace = comp_name
            true_cppname = '"%s::%s"' % (comp_name, mdef.get_name())
            enumr  = "\n"
            enumr += 'cdef extern from "<%s>" namespace "%s":\n' % (internal_file_name, namespace)
            enumr += "\n"
            enumr += '    cdef enum %s %s:\n' % (mdef.get_name(), true_cppname)
            for val in mdef.get_enumvalue():
                enumr += "        %s\n" % val.get_name()
            tres.setMessage(tres.getMessage() + enumr)
        elif len(klass[0].items) != len(mdef.get_enumvalue()):
            tres.setPassed(False)
            tres.setMessage("TODO: Found enum in C++ with %s members but in Cython there are %s members: " % (
                len(mdef.get_enumvalue()), len(klass[0].items) ) )
        else:
            tres.setPassed(True)

    elif kind == "variable" and protection == "public":
        attrnames = [a.name for a in pxd_class.attributes]
        cnt.public_variables += 1
        if not mdef.name in attrnames:
            tres.setPassed(False)
            tres.setMessage("TODO: Found attribute in C++ but not in pxd: %s %s %s" % (mdef.kind, mdef.prot, mdef.name) )
            cnt.public_variables_missing += 1
        else:
            tres.setPassed(True)

    elif kind == "function" and protection == "public":
        # Wrap of public member functions ...
        cnt.public_methods += 1
        c_return_type = mdef.resolve_return_type()
        if mdef.name in pxd_class.methods:
            # Found match between C++ method and Python method
            py_methods = pxd_class.methods[mdef.name]
            if not isinstance(py_methods, list):
                py_methods = [py_methods]

            py_return_type = [str(d.result_type) for d in py_methods]
            if mdef.definition == mdef.name:
                # Constructor, no return type -> all is good
                if len(c_return_type) != 0:
                    raise AssertionError()
                tres.setPassed(True)
            elif "void" in py_return_type and not "void" in c_return_type:
                tres.setPassed(False)
                tres.setMessage( "TODO: Mismatch between C++ return type (%s) and Python return type (%s) in %s %s %s:" % (
                     str(c_return_type), str(py_return_type), mdef.kind, mdef.prot, mdef.name) )
            else:
                tres.setPassed(True)

        else:

            # Missing method, lets remove false positives (destructors, operators, etc)
            cnt.public_methods_missing += 1
            if mdef.name.find("~") != -1:
                # destructor
                cnt.public_methods_missing_nowrapping += 1
                tres.setPassed(True)
                tres.setMessage("Cannot wrap destructor")
            elif mdef.definition == mdef.name:
                # constructor
                find_match = False
                for kk in pxd_class.methods:
                    if kk.split("_")[-1] == mdef.name:
                        find_match = True
                if find_match:
                    cnt.public_methods_missing_nowrapping += 1
                    tres.setPassed(True)
                    tres.setMessage("Renamed constructor")
                else:
                    tres.setPassed(False)
                    tres.setMessage(" -- TODO missing constructor in PXD: %s nogil except +" % mdef.format_definition_for_cython())

            elif (mdef.name.find("operator") != -1 or
                  mdef.name.find("begin") != -1 or
                  mdef.name.find("end") != -1):
                cnt.public_methods_missing_nowrapping += 1
                tres.setPassed(True)
                tres.setMessage("Cannot wrap method with iterator/operator %s" % mdef.name)
            else:
                tres.setPassed(False)
                tres.setMessage(" -- TODO missing function in PXD: %s nogil except +" % mdef.format_definition_for_cython())
    else:
        # It is neither public function/enum/variable
        tres.setPassed(True)

    # Return the testresult
    return tres
#
## Class for counting occurances
#
class Counter(object):
    def __init__(self):
        self.total = 0
        self.skipped = 0
        self.skipped_could_not_parse = 0
        self.skipped_ignored = 0
        self.skipped_protected = 0
        self.skipped_no_location = 0
        self.skipped_no_sections = 0
        self.skipped_no_pxd_file = 0
        self.skipped_no_pxd_match = 0
        self.parsed = 0
        #
        self.public_enums_total = 0
        self.public_enums_missing = 0
        #
        self.public_methods = 0
        self.public_methods_missing = 0
        self.public_methods_missing_nowrapping = 0
        self.public_variables = 0
        self.public_variables_missing = 0

    def computed_skipped(self):
        self.skipped = self.skipped_could_not_parse +\
            self.skipped_ignored + \
            self.skipped_protected + \
            self.skipped_no_location + \
            self.skipped_no_sections + \
            self.skipped_no_pxd_file + \
            self.skipped_no_pxd_match

    def print_skipping_reason(self):
        self.computed_skipped()
        print ("Skipped files: %s" % self.skipped)
        print ("- Could not parse xml: %s" % self.skipped_could_not_parse)
        print ("- Could not parse location in xml: %s" % self.skipped_no_location)
        print ("- Ignored per ignore-file: %s" % self.skipped_ignored)
        print ("- Protected Compound: %s" % self.skipped_protected)
        print ("- Could not find sections in xml: %s" % self.skipped_no_sections)
        print ("- Could not find associated pxd file : %s" % self.skipped_no_pxd_file)
        print ("- Could not find matching class in pxd file : %s" % self.skipped_no_pxd_match)

    def print_stats(self):
        self.computed_skipped()
        print ("Total files: %s" % self.total)
        print ("Skipped files: %s" % self.skipped)
        print ("Parsed files: %s" % self.parsed)
        print ("Parsed public methods %s (of which were missing %s and %s were operator/destructors) " % (self.public_methods, self.public_methods_missing, self.public_methods_missing_nowrapping))
        print ("  - wrapped %s " % (self.public_methods - self.public_methods_missing))
        print ("  - unwrapped operators/destructors %s " % (self.public_methods_missing_nowrapping))
        print ("  - unwrapped methods %s " % (self.public_methods_missing - self.public_methods_missing_nowrapping))
        print ("Parsed public enums %s (of which were missing %s) " % (self.public_enums_total, self.public_enums_missing))
        print ("Parsed public attributes %s (of which were missing %s) " % (self.public_variables, self.public_variables_missing))
        print ("Note that this script counts each method name only once and only maps from \n"+ \
              "C++ to Python (not the other way around), thus the numbers are slightly inaccurate.")

#
## Class for an OpenMS .h file
#
class OpenMSSourceFile(object):
    """
    Class for an OpenMS .h file

    Can parse out information on current maintainer stored in OpenMS-specific
    format.
    """

    def __init__(self, fname):
        self.fname = fname

    def getMaintainer(self):
        try:
            return self._getMaintainer()
        except IOError:
            return None

    def _getMaintainer(self):
        """
        // $Maintainer: xxx $
        // $Authors: xxx $
        """
        maintainer_reg = re.compile(".*\$\s*Maintainer:([^\$]*)\$")
        with open(self.fname) as f:
            data = f.read()
            maintainer = maintainer_reg.search(data)
            if maintainer is not None:
                return maintainer.group(1).strip()
            else:
                return None


#
## Class holding Doxygen XML file and next one function declaration
#
class DoxygenXMLFile(object):
    """
    The doxygen XML file

    Abstracts the parsing of the Doxygen XML file and contains some reasoning
    about the class (e.g. its members, is it pure abstract, etc.)

    Can generate a viable PXD file from the doxygen information alone.
    """

    def __init__(self, fname):
        self.fname = fname
        self.parsed_file = None
        self.compound = None
        self.parsing_error = False
        self.parsing_error_message = None

    def parse_doxygen(self):
        try:
            self.parsed_file = doxygen_parse(self.fname)
            self.compound = self.parsed_file.get_compounddef()
            return self.parsed_file
        except Exception as e:
            print ("Error parsing doxygen xml file", e.message)
            self.parsing_error_message = e.message
            self.parsing_error = True
            return None

    def getInternalFileName(self):
        location = self.parsed_file.get_compounddef().get_location()
        if location is None:
            return None
        return location.get_file()

    def getCompoundFileLocation(self, source_dir):
        location = self.parsed_file.get_compounddef().get_location()
        if location is None:
            return None
        return os.path.realpath( os.path.join(source_dir, "src", "openms", "include", location.get_file()) )

    def isEmpty(self, discard_defines=False):
        compound = self.compound
        if not discard_defines:
            return len(compound.get_sectiondef()) == 0

        # check whether there is more than defines and typdefs
        empty = True
        for mdef in self.iterMemberDef():
            if not mdef.get_kind() in ["define", "typedef", "slot", "signal"]:
                # DoxMemberKind: define property event variable typedef enum function signal prototype friend dcop slot
                empty = False
        if empty and not len(compound.get_sectiondef()) == 0:
            # contains only typedefs etc
            pass
        return empty

    def get_pxd_from_class(self, dfile, internal_file_name, xml_output_path):
        """
        Generate a viable PXD file
        """

        compound = dfile.compound
        comp_name = compound.get_compoundname()

        #
        # Step 1: generate cimport includes
        #
        includes = ""
        if len(compound.get_includes()) == 1:
            try:
                reffile = os.path.join(xml_output_path, compound.get_includes()[0].get_refid() + ".xml")
                # read includes from ref file
                dreffile = DoxygenXMLFile(reffile).parse_doxygen()
                include_compound = dreffile.get_compounddef()
            except Exception as e:
                print ("Error: Could not read includes from file for compound %s with error %s" % (comp_name, e.message))
                include_compound = compound
        else:
            include_compound = compound

        for inc in include_compound.get_includes():
            val = inc.getValueOf_()
            if val.startswith("OpenMS"):
                header_file = val.split("/")[-1]
                if header_file.endswith(".h"):
                    header_file = header_file[:-2]
                # We do not import certain headers since we did not wrap them in Python
                if header_file in ["Exception", "Macros", "config", "StandardTypes"]:
                    continue
                includes += "from %s cimport *\n" % header_file

        #
        # Step 2: class definition
        #
        parent_classes = [n.getValueOf_() for n in compound.basecompoundref]
        namespace = "::".join(comp_name.split("::")[:-1])
        preferred_classname = "_".join(comp_name.split("::")[1:])
        preferred_classname = comp_name.split("::")[-1]
        cldef  = "\n"
        cldef += 'cdef extern from "<%s>" namespace "%s":\n' % (internal_file_name, namespace)
        cldef += "\n"

        inherit_txt = ""
        true_cppname = '"%s"' % comp_name
        if len(parent_classes) > 0:
            inherit_txt = "(%s)" % ",".join(parent_classes)
            true_cppname = '' # Cython does not accept this if mixed with inheritance

        if compound.templateparamlist is None:
            cldef += '    cdef cppclass %s%s %s:\n' % (preferred_classname, inherit_txt, true_cppname)
        else:
            targs = [p.get_declname() for p in compound.templateparamlist.get_param()]
            cldef += '    cdef cppclass %s[%s]%s:\n' % (preferred_classname, ",".join(targs), inherit_txt)
        cldef += '        #\n'
        cldef += '        # wrap-doc:\n'
        cldef += '        #     ADD PYTHON DOCUMENTATION HERE\n'
        cldef += '        #\n'
        if len(parent_classes) > 0:
            cldef += '        # wrap-inherits:\n'
        for p in parent_classes:
            cldef += '        #  %s\n' % (p)

        # check if it is abstract, then do not attempt to wrap it
        if self.isAbstract():
            cldef += '        # wrap-ignore\n'
            cldef += '        # ABSTRACT class\n'

        #
        # Step 3: methods and enums
        #
        methods = ""
        default_ctor = False
        copy_ctor = False
        enum = ""
        static_methods = ""
        imports_needed = {}
        for mdef in dfile.iterMemberDef():
            if mdef.kind == "enum" and mdef.prot == "public":
                # add enums
                enum += '\n'
                enum += 'cdef extern from "<%s>" namespace "%s":\n' % (internal_file_name, comp_name)
                enum += '    cdef enum %s "%s":\n' % (mdef.name, comp_name + "::" + mdef.name)
                enum += '        #wrap-attach:\n'
                enum += '        #    %s\n' % preferred_classname
                for val in mdef.enumvalue:
                    enum += '        %s\n' % val.get_name()

            if mdef.kind == "variable" and mdef.prot == "public":
                # print ("var", mdef.name)
                # cannot wrap const member variables
                if mdef.definition.find("const") == -1:
                    methods += "        %s\n" % mdef.format_definition_for_cython(False)
                else:
                    methods += "        # const # %s\n" % mdef.format_definition_for_cython(False)
            elif mdef.kind == "function" and mdef.prot == "public":
                if mdef.definition == mdef.name:
                    # Means we have a constructor
                    if mdef.get_argsstring().strip() == "()":
                        # Default constructor
                        default_ctor = True
                        continue
                    elif mdef.get_argsstring().strip().find(mdef.name) != -1 and \
                         mdef.get_argsstring().strip().find(",") == -1:
                        # Copy constructor
                        copy_ctor = True
                        continue
                if mdef.name.find("~") != -1:
                    # Destructor
                    continue
                # res += "do member function/attribute : ", mdef.kind,  mdef.prot, mdef.name
                declaration = mdef.format_definition_for_cython(False)
                DoxygenCppFunction.compute_imports(declaration, imports_needed)
                if declaration.find("operator=(") != -1:
                    # assignment operator, cannot be overriden in Python
                    continue
                if mdef.definition.find("static") != -1:
                    methods += "        # TODO: static # %s nogil except +\n" % declaration
                    static_methods += "        %s nogil except + # wrap-attach:%s\n" % (declaration, preferred_classname)
                    continue
                methods += "        %s nogil except +\n" % declaration

        # Build up the whole file
        res  = DoxygenCppFunction.generate_imports(imports_needed) # add default cimport
        res += includes
        res += cldef
        # We need to create a default ctor in any case, however we do not need
        # to *wrap* the copy constructor even though we need to have one for Cython
        if True: # not default_ctor:
            res += "        %s() nogil except +\n" % comp_name.split("::")[-1]
        if not copy_ctor:
            res += "        %s(%s) nogil except + #wrap-ignore\n" % (comp_name.split("::")[-1], comp_name.split("::")[-1])
        else:
            res += "        %s(%s) nogil except +\n" % (comp_name.split("::")[-1], comp_name.split("::")[-1])
        res += methods
        res += enum
        res += "\n"
        if len(static_methods) > 0:
            res += "\n"
            res += "# COMMENT: wrap static methods\n"
            res += 'cdef extern from "<%s>" namespace "%s::%s":\n' % (internal_file_name, namespace, preferred_classname)
            res += "\n"
            res += static_methods
            res += "\n"

        return res

    def iterMemberDef(self):
        """Iterate over all members of this class.

        We do not care about the sections defined in the documentation here.
        """
        for sdef in self.compound.get_sectiondef():
            for mdef_ in sdef.get_memberdef():
                mdef = DoxygenCppFunction.generate_from_obj(mdef_)
                mdef.parent_doxy_file = self
                mdef.compoundname = self.compound.compoundname
                yield mdef

    def isAbstract(self):
        for mdef in self.iterMemberDef():
                if mdef.get_argsstring().endswith("=0"):
                    return True
        return False

class DoxygenCppFunction(object):
    """ A Cpp function definition from a doxygen file"""
    def __init__(self):
        # No other code here, below is the real init method!
        self.initialize_dgencpp()

    @staticmethod
    def generate_from_obj(mdef):
        """Attaches the functionality of this object to the given input object"""
        for k,v in DoxygenCppFunction.__dict__.items():
            if callable(v) and not k == "__init__":
                import types
                mdef.__dict__[k] = types.MethodType(v, mdef)
        mdef.initialize_dgencpp()
        return mdef

    @staticmethod
    def generate_imports(imports):
        """
        Generate default imports
        """
        res = ""
        res += "from Types cimport *\n"
        for k in sorted(imports.keys()):
            if k == "bool":
                res += "from libcpp cimport bool\n"
            else:
                res += "from libcpp.%s cimport %s as libcpp_%s\n" % (k,k,k)
        return res

    @staticmethod
    def compute_imports(declaration, imports):
        if declaration.find("libcpp_vector") != -1:
            imports["vector"] = 0
        if declaration.find("libcpp_pair") != -1:
            imports["pair"] = 0
        if declaration.find("libcpp_map") != -1:
            imports["map"] = 0
        if declaration.find("libcpp_set") != -1:
            imports["set"] = 0
        if declaration.find("libcpp_string") != -1:
            imports["string"] = 0
        if declaration.find("bool") != -1:
            imports["bool"] = 0

    def initialize_dgencpp(self):
        pass

    def resolve_return_type(self):
        res = []
        return self._resolve_type(self.get_type().content_)

    def _resolve_type(self, mtype):
        res = []
        for c in mtype:
            val = c.getValue()
            if hasattr(val, "getValueOf_"):
                res.append(val.getValueOf_())
            else:
                res.append(val)
        return res

    def format_definition_for_cython(self, replace_nogil=True):
        """Parse a doxygen function definition and write it in Cython"""
        c_return_type = self.resolve_return_type()

        # remove default arguments, Cython doesnt like them
        arguments = re.sub("\=[^,\)]*", "", self.get_argsstring())
        function_name = self.name

        arguments = "("
        nested = False
        for i,p in enumerate(self.get_param()):
            ptype = self._resolve_type(p.get_type().content_)
            dname = p.declname
            # ignore default arguments etc ... Cython cannot use them
            # p.defval.content_
            # replace python keywords in argument name: except, type, lamdba, map ...
            import keyword
            if keyword.iskeyword(dname):
                dname = dname + "_"
            if dname in dir(__builtins__):
                dname = dname + "_"
            # dname = dname.replace("except", "except_").replace("type", "type_").replace("lambda", "lambda_").replace("map", "map_")
            tojoin = "".join(ptype) + " " + dname.strip()
            if tojoin.count("std::") > 2:
                nested = True
            if i == 0:
                arguments += tojoin
            else:
                arguments += ", " + "".join(ptype) + " " + dname.strip()

        arguments += ")"
        # arguments = (std::vector<  Int  > column_indices, std::vector<  DoubleReal  > column_values, const  String  & name)

        if len(self.get_argsstring()) == 0:
            arguments = ""

        # remove returned references and const values (Cython cannot deal with those at the moment)
        return_type = "".join(c_return_type)
        return_type = return_type.replace("&", "")
        return_type = return_type.replace("const", "")
        return_type = return_type.strip()
        cpp_def = return_type + " " + function_name + arguments

        # Handle comments
        cpp_def = cpp_def.replace("///", "#")
        cpp_def = cpp_def.replace("//", "#")

        # Add nogil
        if replace_nogil:
            cpp_def = cpp_def.replace(";", "nogil except +")
            cpp_def = cpp_def.replace("const;", "nogil except +")
        else:
            cpp_def = cpp_def.replace("const;", "")
            cpp_def = cpp_def.replace(";", "")

        # Replace common names from OpenMS, templates, STL constructs etc
        # TODO handle static ...
        cpp_def = cpp_def.replace("static", "")
        cpp_def = cpp_def.replace("MSSpectrum<>", "MSSpectrum")
        cpp_def = cpp_def.replace("MSChromatogram<>", "MSChromatogram")
        cpp_def = cpp_def.replace("std::vector", "libcpp_vector")
        cpp_def = cpp_def.replace("std::map", "libcpp_map")
        cpp_def = cpp_def.replace("std::pair", "libcpp_pair")
        cpp_def = cpp_def.replace("std::set", "libcpp_set")
        cpp_def = cpp_def.replace("std::string", "libcpp_string")
        cpp_def = cpp_def.replace("<", "[")
        cpp_def = cpp_def.replace(">", "]")
        cpp_def = cpp_def.replace("operator[]", "operator__[]")
        cpp_def = cpp_def.replace("operator]", "operator>")
        cpp_def = cpp_def.replace("operator[", "operator<")
        cpp_def = cpp_def.replace("operator__[]", "operator[]")

        # Note that template arguments cannot be typedefs but need to be basic types
        cpp_def = cpp_def.replace("[ DoubleReal ]", "[ double ]")
        cpp_def = cpp_def.replace("[ Size ]", "[ size_t ]")
        cpp_def = cpp_def.replace("[Size,Size]", "[size_t,size_t]")
        cpp_def = cpp_def.replace("[ Int ]", "[ int ]")

        cpp_def = cpp_def.replace("FeatureMap[]", "FeatureMap[Feature]")
        cpp_def = cpp_def.replace("MSSpectrum[]", "MSSpectrum")
        cpp_def = cpp_def.replace("MSExperiment[]", "MSExperiment")
        cpp_def = cpp_def.replace("PeakSpectrum", "MSSpectrum")
        cpp_def = cpp_def.replace("PeakMap", "MSExperiment")

        # Handle const
        cpp_def = cpp_def.replace("const String", "constXXXString")
        cpp_def = cpp_def.replace("const ", "")
        cpp_def = cpp_def.replace("constXXXString", "const String")

        # Alert the user to potential problems and comment out potential
        # dangerous things (raw pointers, iterators)
        if cpp_def.find("*") != -1 or \
           cpp_def.find("::iterator") != -1:
            cpp_def = "# POINTER # " + cpp_def
        if cpp_def.find("::") != -1:
            cpp_def = "# NAMESPACE # " + cpp_def
        if self.templateparamlist is not None:
            cpp_def = "# TEMPLATE # " + cpp_def
        if nested:
            cpp_def = "# NESTED STL # " + cpp_def

        return cpp_def.strip()


#
## Class for the ignore file
#
class IgnoreFile(object):
    """
    Describes the ignore file (e.g. which classes we should skip)
    """

    def __init__(self):
        self.data = {
            "IgnoreNames" : [],
            "IgnoreMethods" : {},
        }

    def load(self, fname):
        self.fname = fname
        self.data = yaml.load(open(self.fname) )["PyOpenMSChecker"]

    def isNameIgnored(self, name):
        return name in self.data["IgnoreNames"]

    def getIgnoredMethods(self, name):
        res = self.data["IgnoreMethods"].get(name, [])
        if res is None:
            return []
        return res

class PXDFileParseError(Exception):
    def __init__(self, message):
        self.message = message

#
## Class for the .pxd file
#
class PXDFile(object):
    """
    Class for the .pxd file
    """

    def __init__(self):
        pass

    @staticmethod
    def parse_multiple_files(pxdfiles, comp_name):

        def cimport(b, _, __):
            print ("cimport", b.module_name, "as", b.as_name)

        handlers = { CEnumDefNode : EnumDecl.parseTree,
                     CppClassNode : CppClassDecl.parseTree,
                     CTypeDefNode : CTypeDefDecl.parseTree,
                     CVarDefNode  : MethodOrAttributeDecl.parseTree,
                     CImportStatNode  : cimport,
                     }

        found = False
        # Go through all files and all classes in those files, trying to find
        # the class whose C/C++ name matches the current compound name
        for pxdfile in pxdfiles:
            cython_file = parse_pxd_file(pxdfile)
            for klass in cython_file:
                if hasattr(klass[0], "cname"):
                    if klass[0].cname == comp_name:
                        found = True
                if found: break
            if found: break

        if not found:
            error_str = "Could not find a match for class %s in file %s" % (comp_name, pxdfile)
            raise PXDFileParseError(error_str)

        # Check if we really have a class, then initialize it
        if isinstance(klass[0], CppClassNode):
            cl = CppClassDecl.parseTree(klass[0], klass[1], klass[2])
        else:
            print ("Something is wrong, not a class")
            raise PXDFileParseError("wrong")

        cl.pxdfile = pxdfile
        for klass in cython_file:
            handler = handlers.get(type(klass[0]))
            res = handler(klass[0], klass[1], klass[2])
            if "wrap-attach" in res.annotations:
                if res.annotations["wrap-attach"] == cl.name:
                    ## attach this to the above class
                    cl.methods[res.name] = res

        return cl

class TestResult:
    """ A Result from a single test which either passed or failed.

    In addition, it contains information about the reason why the test failed,
    who is the maintainer and a unique testname.
    """

    def __init__(self, passed = None, message=None, log_level=None, name=None):
        self.message = message
        self.passed = passed
        self.name = name
        self.log_level = log_level
        self.maintainer = None
        if log_level is None:
            self.log_level = 0

    def setMessage(self, message_):
        self.message = message_

    def getMessage(self):
        return self.message

    def getXMLName(self):
        xmlname = self.name
        xmlname = xmlname.replace("::", "_")
        xmlname = re.sub('[^0-9A-Za-z_]', '', xmlname)
        xmlname = xml_escape(xmlname) # still escape, just to be sure
        return xmlname

    def setPassed(self, passed):
        self.passed = passed

    def isPassed(self):
        return self.passed

    def setMaintainer(self, maintainer):
        self.maintainer = maintainer

    def getMaintainer(self):
        if self.maintainer is None:
            return "Nobody"
        return self.maintainer

class TestResultHandler:
    """ A Container for all test results.

    """

    def __init__(self):
        self._list = []

    def append(self, l):
        self._list.append(l)

    def __iter__(self):
        for l in self._list:
            yield l

    def to_cdash_xml(self, template_path, output_path):

        if template_path.endswith("Test.xml"):
            body_start = "<Testing>"
        elif template_path.endswith("Build.xml"):
            body_start = "<Build>"
        else:
            raise Exception("Unsupported template name %s" % template_path)
        xml_output = []
        # load template head (everything up to "<Testing>")
        # -> this assumes a specific format of the xml
        with open(template_path) as f:
            for line in f:
                if line.strip() == body_start:
                    break
                xml_output.append(line)

        """
          # load template head
          $template = file($ctestReportingPath."/Test.xml");
          $newTestFile = array();
          foreach ($template as $line)
          {
            array_push($newTestFile, $line);
            if (trim($line) == "<Testing>")
            {
              break;
            }
          }
          """

        # Start writing the xml
        xml_output.append("  <Testing>\n")
        xml_output.append("  <StartDateTime>%s</StartDateTime>\n" % (time.strftime('%b %d %H:%M') ) )
        xml_output.append("  <StartTestTime>%s</StartTestTime>\n" % (time.time()) )
        xml_output.append("  <TestList>\n")
        for classtestresults in self:
            for tres in classtestresults:
                xml_output.append("    <Test>%s</Test>\n" % xml_escape(tres.getXMLName() )  )
        xml_output.append("  </TestList>\n")

        for classtestresults in self:
            for tres in classtestresults:
                status = ""
                if tres.isPassed():
                    status = "passed"
                else:
                    status = "failed"

                xml_output.append(" " * 2 + '<Test Status="%s">\n' % status)
                xml_output.append(" " * 4 + '<Name>%s</Name>\n' % xml_escape(tres.getXMLName() ) )
                xml_output.append(" " * 4 + '<Path> ./tools/ </Path>\n' )
                xml_output.append(" " * 4 + '<FullName>%s</FullName>\n' % xml_escape(tres.name) )
                xml_output.append(" " * 4 + '<FullCommandLine>python PythonExtensionChecker.py %s</FullCommandLine>\n' % xml_escape(tres.name) )
                xml_output.append(" " * 4 + '<Results>')
                xml_output.append("""
      <NamedMeasurement type="numeric/double" name="Execution Time"><Value>0.001</Value></NamedMeasurement>
      <NamedMeasurement type="text/string" name="Completion Status"><Value>Completed</Value></NamedMeasurement>
      <NamedMeasurement type="text/string" name="Maintainer"><Value>%s</Value></NamedMeasurement>
      <NamedMeasurement type="text/string" name="Command Line"><Value>python PythonExtensionChecker.py</Value></NamedMeasurement>\n""" % (
          xml_escape(tres.getMaintainer() )
      ) )
                xml_output.append(" " * 6 + '<Measurement>\n')
                xml_output.append(" " * 8 + '<Value>\n')
                if not tres.getMessage() is None:
                    xml_output.append(" " * 10 +  xml_escape(tres.getMessage() ) + "\n")
                xml_output.append(" " * 8 + '</Value>\n')
                xml_output.append(" " * 6 + '</Measurement>\n')

                xml_output.append(" " * 4 + '</Results>\n')
                xml_output.append(" " * 2 + '</Test>\n')


        xml_output.append("<EndDateTime>%s</EndDateTime>\n" % (time.strftime('%b %d %H:%M') ) )
        xml_output.append("<EndTestTime>%s</EndTestTime>\n" % (time.time()) )
        xml_output.append("<ElapsedMinutes></ElapsedMinutes>\n")
        xml_output.append("</Testing>\n")
        xml_output.append("</Site>\n")

        with open(output_path, "w") as f:
            for line in xml_output:
                f.write(line)

        """
          /*
          <?xml version="1.0" encoding="UTF-8"?>
          <Site BuildName="Darwin-clang++"
                  BuildStamp="20121021-2300-Nightly"
                  Name="laphroaig.imp.fu-berlin.de"
                  Generator="ctest-2.8.9"
                  CompilerName="/usr/bin/clang++"
                  OSName="Mac OS X"
                  Hostname="laphroaig.imp.fu-berlin.de"
                  OSRelease="10.7.5"
                  OSVersion="11G63"
                  OSPlatform="x86_64"
                  Is64Bits="1"
                  VendorString="GenuineIntel"
                  VendorID="Intel Corporation"
                  FamilyID="6"
                  ModelID="37"
                  ProcessorCacheSize="32768"
                  NumberOfLogicalCPU="4"
                  NumberOfPhysicalCPU="2"
                  TotalVirtualMemory="512"
                  TotalPhysicalMemory="8192"
                  LogicalProcessorsPerPhysical="8"
                  ProcessorClockFrequency="2660"
          >
          <Testing>
          <StartDateTime>Oct 22 18:36 CEST</StartDateTime>
          <StartTestTime>1350923805</StartTestTime>
          <TestList>
            <Test>BinaryComposeFunctionAdapter_test</Test>
          </TestList>
          <Test Status="passed">
            <Name>BinaryComposeFunctionAdapter_test</Name>
            <Path>./source/TEST</Path>
            <FullName>./source/TEST/BinaryComposeFunctionAdapter_test</FullName>
            <FullCommandLine>/Users/aiche/dev/openms/openms-src/build/ninja/source/TEST/bin/BinaryComposeFunctionAdapter_test</FullCommandLine>
            <Results>
                    <NamedMeasurement type="numeric/double" name="Execution Time"><Value>0.469694</Value></NamedMeasurement>
                    <NamedMeasurement type="text/string" name="Completion Status"><Value>Completed</Value></NamedMeasurement>
                    <NamedMeasurement type="text/string" name="Command Line"><Value>/Users/aiche/dev/openms/openms-src/build/ninja/source/TEST/bin/BinaryComposeFunctionAdapter_test</Value></NamedMeasurement>
                    <Measurement>
                      <Value>
                      freier Text
                      </Value>
                    </Measurement>
            </Results>
          </Test>
              <EndDateTime>Oct 22 18:43 CEST</EndDateTime>
              <EndTestTime>1350924239</EndTestTime>
          <ElapsedMinutes>7.2</ElapsedMinutes></Testing>
        </Site>
          */
                """

def writeOutput(testresults, output_format, cnt, bin_path):
    ###################################
    #   Output
    ###################################
    if output_format in ["text", "text-verbose", "text-quiet"]:
        for classtestresults in testresults:
            if len(classtestresults) > 1:
                t = classtestresults[0]
                lenfailed = len([t for t in classtestresults if not t.isPassed() ] )
                if lenfailed > 0:
                    print ("== Test results for element %s - from Cpp file %s with maintainer %s and corresponding pxd file %s" % (
                        t.comp_name, t.file_location, t.maintainer, t.pxdfile))

            for tres in classtestresults:
                if not tres.isPassed():
                    print (tres.message)
                elif tres.log_level >= 10 and output_format in ["text", "text-verbose"]:
                    print (tres.message)
                elif tres.log_level >= 0 and output_format in ["text-verbose"]:
                    print (tres.name, "::", tres.message)

    elif output_format == "xml":

        # check if all files required to report in CDash are present
        tag_file = os.path.join(bin_path, "Testing", "TAG" )
        try:
            # read the first line of tagfile (TAG) -> if it does not exist,
            # an IOError is thrown
            with open(tag_file) as f:
                ctestReportingPath = f.readline().strip()
                ctestReportingPath = os.path.join(bin_path, "Testing", ctestReportingPath)
                if not os.path.exists( ctestReportingPath ):
                    raise Exception("Missing directory at %s" % ( ctestReportingPath ) )
        except IOError:
            raise Exception("Missing nightly test information at %s" % (tag_file) )

        template_path = os.path.join(ctestReportingPath, "Test.xml" )
        output_path = template_path # output is always Test.xml
        if not os.path.isfile(template_path):
            template_path = os.path.join(ctestReportingPath, "Build.xml" ) #Build.xml an be used as alternative template

        testresults.to_cdash_xml(template_path, output_path)

    else:
        raise Exception("Unknown output format %s" % output_format)

    cnt.print_stats()
    cnt.print_skipping_reason()


def checkPythonPxdHeader(src_path, bin_path, ignorefilename, pxds_out, print_pxd, output_format, generate_pxd, verbose):
    """ Checks a set of doxygen xml file against a set of pxd header files

    For each C++ class found in the doxygen XML files, it tries to identify the
    corresponding pxd file. If a pxd file exists, it checks whether

    i)   all public functions, enums and attributes are wrapped in Python
    ii)  all void return types are correct in Python (these are not checked at
         compile time)
    iii) all fields of an enum are accessible from Python

    If it finds a method missing, the script suggests an addition and if a
    whole class is missing, the script writes suggestion .pxd file to a
    specified location (pxds_out).

    The output format can either be in text form (human readable) or in xml
    form which will try to overwrite the cdash Test.xml file to proivide an
    output to cdash. Please only specify xml output if in your binary
    you have executed "ctest -D Nightly" or similar.

    TODO also look at ./doc/doxygen/doxygen-error.log ?

    Make sure to build the doxygen xmls first with
    $ make doc_xml

    """

    xml_output_path = os.path.join(bin_path, "doc", "xml_output")
    xml_files = glob.glob(xml_output_path + "/*.xml")
    print ("Found %s doxygen xml files" % (len(xml_files)))
    if len(xml_files) == 0:
        raise Exception("No doxygen files found in directory:\n%s,\n" % xml_output_path + \
                        "Please make sure you build the doxygen xmls (make doc_xml)\n" +\
                        "and that you specified the correct directory." )

    print ("Creating pxd file map")
    pxd_file_matching = create_pxd_file_map(src_path)
    print ("Found %s matching pxd files" % len(pxd_file_matching))
    cnt = Counter()
    cnt.total = len(xml_files)
    ignorefile = IgnoreFile()
    if len(ignorefilename) > 0:
        ignorefile.load(ignorefilename)

    if len(generate_pxd) > 0:
        print ("Will only consider class", generate_pxd)

    def pxd_text_printout(pxd_text, pxds_out, comp_name, print_pxd):
        if print_pxd:
            print ("")
            print (pxd_text)
        if len(pxds_out) > 0 and pxd_text is not None:
            fname = os.path.join(pxds_out, "%s.pxd" % comp_name.split("::")[-1] )
            with open(fname, "w" ) as f:
                f.write(pxd_text)

    testresults = TestResultHandler()

    # Iterate through all xml files generated by doxygen (these are all the
    # classes available in OpenMS)
    for f in xml_files:

        # Only look out for one specific pxd file (see option --generate_pxd_for)
        if len(generate_pxd) > 0:
            if f.find(generate_pxd) == -1:
                continue

        if verbose:
            print ("Working on file", f)

        # Try to parse the doxygen file
        dfile = DoxygenXMLFile(f)
        res = dfile.parse_doxygen()
        if dfile.parsing_error:
            # e.g. <computeroutput><bold></computeroutput>
            cnt.skipped_could_not_parse += 1
            msg = "Skip:: No-parse :: could not parse file %s with error %s" % (f, dfile.parsing_error_message)
            tres = TestResult(False, msg, name="%s_test" % f )
            testresults.append([ tres ])
            if verbose: print ("  - Skip file due to parsing error")
            continue
        elif os.path.basename(f) == "index.xml":
            # Skip the index file
            continue

        # Parse class and namespace
        # Skip certain namespaces or those without any namespace (we are only
        # interested in the OpenMS and OpenSwath namespace).
        compound = res.get_compounddef()
        comp_name = compound.get_compoundname()
        if len(comp_name.split("::") ) == 1:
            # We are only interested in the classes themselves (in OpenMS
            # namespace), we thus skip all TOPP tools, header and cpp
            # descriptors which are not inside a namespace:
            if verbose: print ("  - Skip file without namespace:", comp_name)
            continue

        if verbose:
            print ("  - Found class", comp_name, compound.prot, "in namespace", comp_name.split("::")[0])

        namespace = comp_name.split("::")[0]
        if namespace in ["std", "Ui", "xercesc", "seqan"]:
            # Namespace std, xerces, UI -> skip
            continue
        elif comp_name.startswith("ms::numpress"):
            # MS Numpress namespace
            continue
        elif comp_name.startswith("KDTree::"):
            # KD Tree namespace
            continue
        elif not (comp_name.startswith("OpenMS") or comp_name.startswith("OpenSwath") or comp_name.startswith("RNPxl") ):
            # Continue without checking or generating a testreport
            print ("Unknown namespace", comp_name)
            continue

        # Skip files which are listed in the "ignore" file
        if ignorefile.isNameIgnored(comp_name):
            msg = "Skip:: Ignored :: Class %s (file %s)" % (comp_name, f)
            tres = TestResult(True, msg, log_level=10, name="%s_test" % comp_name)
            testresults.append([ tres ])
            cnt.skipped_ignored += 1
            continue

        # Ignore private/protected classes
        if compound.prot != "public":
            msg = "Skip:: Protected :: Compound %s is not public, skip" % (comp_name)
            tres = TestResult(True, msg, log_level=10, name="%s_test" % comp_name)
            testresults.append([ tres ])
            cnt.skipped_protected += 1
            continue

        # Get file location and skip empty files
        file_location = dfile.getCompoundFileLocation(src_path)
        internal_file_name = dfile.getInternalFileName()
        if verbose:
            print ("  - Header file location identified as", internal_file_name)

        if file_location is None:
            msg = "Skip:: No-data :: there is no source file for %s" % f
            tres = TestResult(True, msg, log_level=10, name="%s_test" % comp_name)
            testresults.append([ tres ])
            cnt.skipped_no_location += 1
            continue

        # Skip empty classes
        openms_file = OpenMSSourceFile(file_location)
        maintainer = openms_file.getMaintainer()
        if dfile.isEmpty(True):
            msg = "Skip:: No-data :: File is empty (no section definitions found or only definitions found) in file %s" % f
            tres = TestResult(True, msg, log_level=10, name="%s_test" % comp_name)
            tres.maintainer = maintainer
            testresults.append([ tres ])
            cnt.skipped_no_sections += 1
            continue

        # Retrieve all associated pxd files with this specific header file
        if internal_file_name in pxd_file_matching:
            pxdfiles = pxd_file_matching[internal_file_name]
        else:
            msg = "Skip:: No-pxd :: No pxd file exists for Class %s (File %s) %s" % (comp_name, file_location, f)
            tres = TestResult(False, msg,  name="Missing_%s_test" % comp_name )
            tres.maintainer = maintainer
            testresults.append([ tres ])
            cnt.skipped_no_pxd_file += 1
            pxd_text = dfile.get_pxd_from_class(dfile, internal_file_name, xml_output_path)
            pxd_text_printout(pxd_text, pxds_out, comp_name, print_pxd)
            continue

        if verbose:
            print ("  - Matching pxd files", pxdfiles)

        # At this point we have
        #  - the cpp class as parsed by Doxygen
        #  - the corresponding OpenMS header file
        #  - the matching pxd file(s) with the Python wrappers

        # Parse the pxd files corresponding to this doxygen XML file
        try:
            # Raise a (dummy) exception to actually produce a PXD file for a
            # specific class if requested by the user.
            if len(generate_pxd) > 0:
                raise PXDFileParseError ("dummy")
            pxd_class = PXDFile.parse_multiple_files(pxdfiles, comp_name)
            pxdfile = pxd_class.pxdfile
        except PXDFileParseError as e:
            # TODO specific exception
            msg = "Skip:: No-pxd :: " + e.message + " for %s (in pxd file %s)" % (comp_name, pxdfiles)
            tres = TestResult(False, msg,  name="Missing_%s_test" % comp_name )
            tres.maintainer = maintainer
            testresults.append([ tres ])
            cnt.skipped_no_pxd_match += 1
            pxd_text = dfile.get_pxd_from_class(dfile, internal_file_name, xml_output_path)
            pxd_text_printout(pxd_text, pxds_out, comp_name, print_pxd)
            continue

        # Count the current file
        cnt.parsed += 1

        # Loop through all methods which are listed in the doxygen XML file and match them to the pxd file
        classtestresults = []
        for method_cntr, mdef in enumerate(dfile.iterMemberDef()):

            if mdef.get_name() in ignorefile.getIgnoredMethods(comp_name):
                msg = "Ignore member function/attribute : %s %s %s " % (mdef.kind, mdef.prot, mdef.name)
                tres = TestResult(True, msg, log_level=10)
            else:
                tres = handle_member_definition(mdef, pxd_class, cnt)

            testname = "%s_%s::%s" % (comp_name, method_cntr, mdef.name)
            testname = testname.replace("::", "_")
            testname = re.sub('[^a-zA-Z0-9_]+', '', testname)
            tres.comp_name = comp_name
            tres.file_location = file_location
            tres.pxdfile = pxdfile
            tres.maintainer = maintainer
            tres.name = testname
            classtestresults.append(tres)

        testresults.append(classtestresults)

    writeOutput(testresults, output_format, cnt, bin_path)

def main(options):
    checkPythonPxdHeader(options.src_path, options.bin_path,
                         options.ignorefile, options.pxds_out,
                         options.print_pxd, options.output_format,
                         options.generate_pxd, options.verbose)

def handle_args():
    usage = "Python extension checker. Run to identify classes and functions that have not been wrapped yet in pyOpenMS. Make sure you run 'make doc_xml' in the build path (--bin_path) first."

    parser = argparse.ArgumentParser(description = usage )
    parser.add_argument("--bin_path", dest="bin_path", default=".", help="OpenMS build path")
    parser.add_argument("--src_path", dest="src_path", default=".", help="OpenMS source path")
    parser.add_argument("--ignore-file", dest="ignorefile", default="", help="Checker ignore file")
    parser.add_argument("--pxds-out", dest="pxds_out", default="", help="Folder to write pxd files")
    parser.add_argument("--generate_pxd_for", dest="generate_pxd", default="", help="Generate pxd file onyl for this class, then exit")
    parser.add_argument("--output", dest="output_format", default="text", help="Output format (valid are 'xml', 'text', 'text-quiet', 'text-verbose' for text or ctest XML format)")
    parser.add_argument('--print_pxd', action='store_true', default=False)
    parser.add_argument('--verbose', action='store_true', default=False, help="Be verbose")
    #   print ("Usage: checker.php <OpenMS src path> <OpenMS build path> [-u \"user name\"] [-t test] [options]\n";)

    args = parser.parse_args(sys.argv[1:])
    return args

if __name__=="__main__":
    options = handle_args()
    main(options)


"""
offending doxygen lines that fail to parse:

include/source/APPLICATIONS/TOPP/IDRipper.cpp
  <B>NOTE: The meta value file origin is removed by the @p IDSplitter!!</B>
  generates
  the <computeroutput>IDSplitter!!</bold></computeroutput>

doc/OpenMS_tutorial/OpenMS_Tutorial.doxygen
  \arg \c <b>[1]</b>:<A HREF="http://bieson.ub.uni-bielefeld.de/frontdoor.php?source_opus=1370">
  generates
  <listitem><para><computeroutput><bold></computeroutput>[1]</bold>:

"""






    def __call__(self, *args):
        assert len(args) in (1, 2), len(args)
        if len(args) == 1:
            return self._score_1(args[0])
        return self._score_2(args[0], args[1])

    def _score_1(self, spec1):
        assert isinstance(spec1, (MSSpectrum,))
        cdef _MSSpectrum * _new_spec1 = new _MSSpectrum()
        cdef _Peak1D _peak

        if True:
            for _peak in deref(<_MSSpectrum *>(<MSSpectrum>spec1).inst.get()):
                _new_spec1.push_back(<_Peak1D>_peak)

        cdef _SpectrumAlignmentScore * scorer = self.inst.get()
        cdef double score = <double>deref(scorer)(deref(_new_spec1))
        del _new_spec1
        return score

    def _score_2(self, spec1, spec2):
        assert isinstance(spec1, (MSSpectrum))
        assert isinstance(spec2, (MSSpectrum))
        cdef _MSSpectrum * _new_spec1 = new _MSSpectrum()
        cdef _MSSpectrum * _new_spec2 = new _MSSpectrum()
        cdef _Peak1D _peak

        if True:
            for _peak in deref(<_MSSpectrum *>(<MSSpectrum>spec1).inst.get()):
                _new_spec1.push_back(<_Peak1D>_peak)

        if True:
            for _peak in deref(<_MSSpectrum *>(<MSSpectrum>spec2).inst.get()):
                _new_spec2.push_back(<_Peak1D>_peak)

        cdef _SpectrumAlignmentScore * scorer = self.inst.get()
        cdef double score = <double>deref(scorer)(deref(_new_spec1), deref(_new_spec2))
        del _new_spec1
        del _new_spec2
        return score



#!/usr/bin/python
# -*- encoding: utf8 -*-
"""Python bindings to the OpenMS C++ library.

The pyOpenMS package contains Python bindings for a large part of the OpenMS
library (http://www.open-ms.de) for mass spectrometry based proteomics. It thus
provides providing facile access to a feature-rich, open-source algorithm
library for mass-spectrometry based proteomics analysis. These Python bindings
allow raw access to the data-structures and algorithms implemented in OpenMS,
specifically those for file access (mzXML, mzML, TraML, mzIdentML among
others), basic signal processing (smoothing, filtering, de-isotoping and
peak-picking) and complex data analysis (including label-free, SILAC, iTRAQ and
SWATH analysis tools).

For further documentation, please see https://pyopenms.readthedocs.io

Please cite:

    Röst HL, Schmitt U, Aebersold R, Malmström L.
    pyOpenMS: a Python-based interface to the OpenMS mass-spectrometry algorithm library.
    Proteomics. 2014 Jan;14(1):74-7. doi: 10.1002/pmic.201300246.

"""
from __future__ import print_function
from __future__ import absolute_import

from .sysinfo import *
from .version import version as __version__

import os
here = os.path.abspath(os.path.dirname(__file__))
os.environ["OPENMS_DATA_PATH"] = os.path.join(here, "share/OpenMS")

import sys
if sys.platform.startswith("linux"):
    # load local shared libries before we import pyopenms.so, else
    # those are not found. setting LD_LIBRARY_PATH does not work,
    # see: http://stackoverflow.com/questions/1178094
    import ctypes
    ctypes.cdll.LoadLibrary(os.path.join(here, "libOpenSwathAlgo.so"))
    ctypes.cdll.LoadLibrary(os.path.join(here, "libOpenMS.so"))
    ctypes.cdll.LoadLibrary(os.path.join(here, "libSuperHirn.so"))

try:
    from .all_modules import *
    from .python_extras import *
except Exception as e:
    print("")
    print("="*70)
    print("Error when loading pyOpenMS libraries!")
    print("Libraries could not be found / could not be loaded.")
    print("")
    print("Note: when using the Spyder IDE, this error may be triggered when")
    print("the 'Automatic' backend is used. Please change this in Tools ->")
    print("Preferences -> IPython -> Graphics to 'Inline'.")
    print("")
    print("To debug this error, please run ldd (on linux) or dependency walker (on windows) on ")
    print("")
    print(os.path.join(here, "pyopenms.so"))
    print("")
    print("="*70)

    try:
        import PyQt4.QtCore
    except:
        pass
    else:
        from .qt_version_info import info

        info = "\n    ".join(info.split("\n"))

        print("""When building pyopenms qmake said:

    %s
PYQT has version %s

Maybe this causes a conflict.  You can test this by importing pyopenms
first and then import PyQt4.QtCore.
        """ % (info, PyQt4.QtCore.PYQT_VERSION_STR) )
    print("="*70)
    print("\n")
    raise e

del os, here, sys



from Types cimport *
from LinearResampler cimport *

cdef extern from "<OpenMS/FILTERING/TRANSFORMERS/LinearResamplerAlign.h>" namespace "OpenMS":
    
    cdef cppclass LinearResamplerAlign(LinearResampler) :
        # wrap-inherits:
        #  LinearResampler
        LinearResamplerAlign(LinearResamplerAlign) nogil except + #wrap-ignore

        # TEMPLATE # void raster(SpecT[ PeakType ] & spectrum) nogil except +
        # TEMPLATE # void raster_align(SpecT[ PeakType ] & spectrum, double start_pos, double end_pos) nogil except +
        # TEMPLATE # void raster(ConstPeakTypeIterator raw_it, ConstPeakTypeIterator raw_end, PeakTypeIterator resample_it, PeakTypeIterator resample_end) nogil except +
        # TEMPLATE # void raster_interpolate(PeakTypeIterator raw_it, PeakTypeIterator raw_end, PeakTypeIterator it, PeakTypeIterator resampled_end) nogil except +




from libcpp cimport bool
from Types cimport *
from String cimport *
from Residue cimport *
from EmpiricalFormula cimport *
from Ribonucleotide cimport *
from Map cimport *

cdef extern from "<OpenMS/CHEMISTRY/NASequence.h>" namespace "OpenMS":

    ctypedef Ribonucleotide RibonucleotideChainEnd

    cdef cppclass NASequence "OpenMS::NASequence":
        # wrap-hash:
        #   toString().c_str()
        #
        # wrap-doc:
        #   Representation of an RNA sequence
        #   This class represents nucleic acid sequences in OpenMS. An NASequence
        #   instance primarily contains a sequence of ribonucleotides. 

        NASequence() nogil except +
        NASequence(NASequence) nogil except + # wrap-ignore

        bool operator==(const NASequence & rhs) nogil except +
        bool operator!=(const NASequence & rhs) nogil except +
        bool operator<(const NASequence & rhs) nogil except +

        libcpp_vector[ const Ribonucleotide * ] getSequence() nogil except +

        const Ribonucleotide * operator[](size_t index) nogil except +

        # check if sequence is empty
        bool empty() nogil except +

        void setSequence(const libcpp_vector[ const Ribonucleotide *] & seq) nogil except +

        # returns the peptide as string with modifications embedded in brackets
        String toString() nogil except +

        # sets the 5' modification
        void setFivePrimeMod(const RibonucleotideChainEnd * modification) nogil except +

        # returns the name (ID) of the N-terminal modification, or an empty string if none is set
        const RibonucleotideChainEnd * getFivePrimeMod() nogil except +

        # sets the 3' modification
        void setThreePrimeMod(const RibonucleotideChainEnd * modification) nogil except +

        const RibonucleotideChainEnd * getThreePrimeMod() nogil except +

        # returns the residue at position index
        const Ribonucleotide * get(Size index) nogil except +

        # set the residue at position index
        void set(size_t index, const Ribonucleotide * r) nogil except +

        # returns the formula of the peptide
        EmpiricalFormula getFormula() nogil except + # wrap-doc:convenience function with ResidueType=Full and charge = 0 by default
        EmpiricalFormula getFormula(NASFragmentType type_, Int charge) nogil except +

        # returns the average weight of the peptide
        double getAverageWeight() nogil except +
        double getAverageWeight(NASFragmentType type_, Int charge) nogil except +

        # returns the mono isotopic weight of the peptide
        double getMonoWeight() nogil except +
        double getMonoWeight(NASFragmentType type_, Int charge) nogil except +

        # returns the number of residues
        Size size() nogil except +

        # returns a peptide sequence of the first index residues
        NASequence getPrefix(Size length) nogil except +

        # returns a peptide sequence of the last index residues
        NASequence getSuffix(Size length) nogil except +

        # returns a peptide sequence of number residues, beginning at position index
        NASequence getSubsequence(Size start, Size length) nogil except +


# COMMENT: wrap static methods
cdef extern from "<OpenMS/CHEMISTRY/NASequence.h>" namespace "OpenMS::NASequence":
        
        
        # static members
        NASequence fromString(const String & s) nogil except +  # wrap-attach:NASequence

cdef extern from "<OpenMS/CHEMISTRY/NASequence.h>" namespace "OpenMS::NASequence":
    cdef enum NASFragmentType "OpenMS::NASequence::NASFragmentType":
        #wrap-attach:
        #    NASequence
        Full
        Internal
        FivePrime
        ThreePrime
        AIon
        BIon
        CIon
        XIon
        YIon
        ZIon
        Precursor
        BIonMinusH20
        YIonMinusH20
        BIonMinusNH3
        YIonMinusNH3
        NonIdentified
        Unannotated
        WIon
        AminusB
        DIon
        SizeOfNASFragmentType




from Types cimport *
from String cimport *
from libcpp cimport bool

cdef extern from "<OpenMS/DATASTRUCTURES/String.h>" namespace "OpenMS":
    cdef cppclass StringView:

        StringView() nogil except +
        StringView(const String& s) nogil except +
        StringView(StringView) nogil except +

        bool operator<(StringView other) nogil except +
        StringView substr(Size start, Size end)  nogil except +
        Size size()  nogil except +
        String getString()  nogil except +



from Types cimport *
from String cimport *
from Feature cimport *
from FeatureMap cimport *
from MSSpectrum cimport *

cdef extern from "<OpenMS/FORMAT/MsInspectFile.h>" namespace "OpenMS":
    
    cdef cppclass MsInspectFile "OpenMS::MsInspectFile":
        MsInspectFile() nogil except +
        MsInspectFile(MsInspectFile) nogil except + #wrap-ignore
        void load(const String & filename, FeatureMap & feature_map) nogil except +
        void store(const String & filename, MSSpectrum & spectrum) nogil except +




from Types cimport *
from libcpp.vector cimport vector as libcpp_vector
from MSExperiment cimport *
from Peak1D cimport *
from ChromatogramPeak cimport *
from String cimport *
from File cimport *
from ProgressLogger cimport *

cdef extern from "<OpenMS/FORMAT/MascotInfile.h>" namespace "OpenMS":
    
    cdef cppclass MascotInfile(ProgressLogger) :
        # wrap-inherits:
        #  ProgressLogger
        MascotInfile() nogil except +
        MascotInfile(MascotInfile) nogil except + #wrap-ignore
        void store(const String & filename, MSSpectrum & spec, double mz, double retention_time, String search_title) nogil except +
        void store(const String & filename, MSExperiment & experiment, String search_title) nogil except +
        void load(const String & filename, MSExperiment & exp) nogil except +
        String  getBoundary() nogil except +
        void setBoundary(const String & boundary) nogil except +
        String  getDB() nogil except +
        void setDB(const String & db) nogil except +
        String  getSearchType() nogil except +
        void setSearchType(const String & search_type) nogil except +
        String  getHits() nogil except +
        void setHits(const String & hits) nogil except +
        String  getCleavage() nogil except +
        void setCleavage(const String & cleavage) nogil except +
        String  getMassType() nogil except +
        void setMassType(const String & mass_type) nogil except +
        libcpp_vector[ String ]  getModifications() nogil except +
        void setModifications(libcpp_vector[ String ] & mods) nogil except +
        libcpp_vector[ String ]  getVariableModifications() nogil except +
        void setVariableModifications(libcpp_vector[ String ] & mods) nogil except +
        String  getInstrument() nogil except +
        void setInstrument(const String & instrument) nogil except +
        UInt getMissedCleavages() nogil except +
        void setMissedCleavages(UInt missed_cleavages) nogil except +
        float getPrecursorMassTolerance() nogil except +
        void setPrecursorMassTolerance(float precursor_mass_tolerance) nogil except +
        float getPeakMassTolerance() nogil except +
        void setPeakMassTolerance(float ion_mass_tolerance) nogil except +
        String  getTaxonomy() nogil except +
        void setTaxonomy(const String & taxonomy) nogil except +
        String  getFormVersion() nogil except +
        void setFormVersion(const String & form_version) nogil except +
        String  getCharges() nogil except +
        void setCharges(libcpp_vector[ int ] & charges) nogil except +




# from MaxLikeliFitter1D cimport *
from String cimport *

cdef extern from "<OpenMS/TRANSFORMATIONS/FEATUREFINDER/IsotopeFitter1D.h>" namespace "OpenMS":
    
    cdef cppclass IsotopeFitter1D "OpenMS::IsotopeFitter1D":
        IsotopeFitter1D() nogil except +
        IsotopeFitter1D(IsotopeFitter1D) nogil except +
        # QualityType fit1d(RawDataArrayType &range, InterpolationModel *&model) nogil except +
        # Fitter1D * create() nogil except +
        String getProductName() nogil except +




from InterpolationModel cimport *
from IsotopeDistribution cimport *
from EmpiricalFormula cimport *

cdef extern from "<OpenMS/TRANSFORMATIONS/FEATUREFINDER/IsotopeModel.h>" namespace "OpenMS":
    
    cdef cppclass IsotopeModel "OpenMS::IsotopeModel":
        IsotopeModel() nogil except +
        IsotopeModel(IsotopeModel) nogil except +
        UInt getCharge() nogil except +
        void setOffset(double offset) nogil except +
        double getOffset() nogil except +
        EmpiricalFormula getFormula() nogil except +
        void setSamples(EmpiricalFormula &formula) nogil except +
        double getCenter() nogil except +
        IsotopeDistribution  getIsotopeDistribution() nogil except +
        # BaseModel[ 1 ] * create() nogil except +
        String getProductName() nogil except +

cdef extern from "<OpenMS/TRANSFORMATIONS/FEATUREFINDER/IsotopeModel.h>" namespace "OpenMS::IsotopeModel":
    cdef enum Averagines "OpenMS::IsotopeModel::Averagines":
        #wrap-attach:
        #    IsotopeModel
        C
        H
        N
        O
        S
        AVERAGINE_NUM




from Types cimport *
from FilterFunctor cimport *
from MSSpectrum cimport *

cdef extern from "<OpenMS/FILTERING/TRANSFORMERS/NeutralLossDiffFilter.h>" namespace "OpenMS":
    
    cdef cppclass NeutralLossDiffFilter(FilterFunctor) :
        # wrap-inherits:
        #  FilterFunctor
        NeutralLossDiffFilter() nogil except +
        NeutralLossDiffFilter(NeutralLossDiffFilter) nogil except +
        double apply(MSSpectrum & ) nogil except +
        # POINTER # FilterFunctor * create() nogil except +
        String getProductName() nogil except +




from Types cimport *
from libcpp cimport bool
from libcpp.map cimport map as libcpp_map
from libcpp.set cimport set as libcpp_set
from libcpp.vector cimport vector as libcpp_vector
from FeatureMap cimport *
from PeptideIdentification cimport *
from PSLPFormulation cimport *
from PrecursorIonSelectionPreprocessing cimport *

cdef extern from "<OpenMS/ANALYSIS/TARGETED/PrecursorIonSelection.h>" namespace "OpenMS":
    
    cdef cppclass PrecursorIonSelection(DefaultParamHandler) :
        # wrap-inherits:
        #  DefaultParamHandler
        PrecursorIonSelection() nogil except +
        PrecursorIonSelection(PrecursorIonSelection) nogil except +

        double  getMaxScore() nogil except +
        void setMaxScore(double & max_score) nogil except +
        void sortByTotalScore(FeatureMap & features) nogil except +
        void getNextPrecursors(FeatureMap & features, FeatureMap & next_features, UInt number) nogil except +

        # TODO immutable types by reference
        # void getNextPrecursorsSeq(FeatureMap & features, FeatureMap & next_features, UInt number, double & rt) nogil except +

        void getNextPrecursors(libcpp_vector[ int ] & solution_indices,
                               libcpp_vector[ IndexTriple ] & variable_indices,
                               libcpp_set[ int ] & measured_variables,
                               FeatureMap & features,
                               FeatureMap & new_features,
                               UInt step_size,
                               PSLPFormulation & ilp) nogil except +
        void rescore(FeatureMap & features, libcpp_vector[ PeptideIdentification ] & new_pep_ids, libcpp_vector[ ProteinIdentification ] & prot_ids, PrecursorIonSelectionPreprocessing & preprocessed_db, bool check_meta_values) nogil except +
        void simulateRun(FeatureMap & features, libcpp_vector[ PeptideIdentification ] & pep_ids, libcpp_vector[ ProteinIdentification ] & prot_ids, PrecursorIonSelectionPreprocessing & preprocessed_db, String path, MSExperiment & experiment, String precursor_path) nogil except +
        void setLPSolver(SOLVER solver) nogil except +
        SOLVER getLPSolver() nogil except +
        void reset() nogil except +
        # TODO not implemented but part of the API
        # libcpp_map[ String, libcpp_set[ String ] ]  getPeptideProteinCounter() nogil except +

cdef extern from "<OpenMS/ANALYSIS/TARGETED/PrecursorIonSelection.h>" namespace "OpenMS::PrecursorIonSelection":
    cdef enum PrecursorIonSelection_Type "OpenMS::PrecursorIonSelection::Type":
        #wrap-attach:
        #    PrecursorIonSelection
        IPS
        ILP_IPS
        SPS
        UPSHIFT
        DOWNSHIFT
        DEX

# cdef extern from "<OpenMS/ANALYSIS/TARGETED/PrecursorIonSelection.h>" namespace "OpenMS::PrecursorIonSelection":
#     
#     cdef cppclass SeqTotalScoreMore :
#         SeqTotalScoreMore(SeqTotalScoreMore) nogil except + #wrap-ignore
#         bool operator()(Feature & left, Feature & right) nogil except + # wrap-cast:evaluate
# 
# cdef extern from "<OpenMS/ANALYSIS/TARGETED/PrecursorIonSelection.h>" namespace "OpenMS::PrecursorIonSelection":
#     
#     cdef cppclass TotalScoreMore :
#         TotalScoreMore(TotalScoreMore) nogil except + #wrap-ignore
#         bool operator()(Feature & left, Feature & right) nogil except + # wrap-cast:evaluate
# 



from libcpp cimport bool
from libcpp.vector cimport vector as libcpp_vector
from DataValue cimport *
from String cimport *
from Types cimport *
from MetaInfoInterface cimport *
from ProteinHit cimport *
from DigestionEnzymeProtein cimport *
from PeptideIdentification cimport *
from DateTime cimport *
# from MSExperiment cimport *

cdef extern from "<OpenMS/METADATA/ProteinIdentification.h>" namespace "OpenMS":

    cdef cppclass ProteinIdentification(MetaInfoInterface):
        # wrap-inherits:
        #    MetaInfoInterface

        ProteinIdentification() nogil except +
        ProteinIdentification(ProteinIdentification) nogil except +

        bool operator==(ProteinIdentification) nogil except +
        bool operator!=(ProteinIdentification) nogil except +

        # Returns the protein hits (mutable)
        libcpp_vector[ProteinHit] getHits() nogil except +
        # Appends a protein hit
        void insertHit(ProteinHit input) nogil except +
        # Sets the protein hits
        void setHits(libcpp_vector[ProteinHit] hits) nogil except +
        # Finds a protein hit by accession (returns past-the-end iterator if not found)
        # libcpp_vector[ProteinHit].iterator findHit(String accession) nogil except +

        # Returns the protein groups (mutable)
        libcpp_vector[ProteinGroup] getProteinGroups() nogil except +
        # Appends a new protein group
        void insertProteinGroup(ProteinGroup group) nogil except +

        # Returns the indistinguishable proteins (mutable)
        libcpp_vector[ProteinGroup] getIndistinguishableProteins() nogil except +
        # Appends new indistinguishable proteins
        void insertIndistinguishableProteins(ProteinGroup group) nogil except +

        # Returns the protein significance threshold value
        double getSignificanceThreshold() nogil except +
        # Sets the protein significance threshold value
        void setSignificanceThreshold(double value) nogil except +
        # Returns the protein score type
        String getScoreType() nogil except +
        # Sets the protein score type
        void setScoreType(String type) nogil except +
        # Returns true if a higher score represents a better score
        bool isHigherScoreBetter() nogil except +
        # Sets the orientation of the score (is higher better?)
        void setHigherScoreBetter(bool higher_is_better) nogil except +
        # Sorts the protein hits according to their score
        void sort() nogil except +
        # Sorts the protein hits by score and assigns ranks (best score has rank 1)
        void assignRanks() nogil except +

        # Compute the coverage (in percent) of all ProteinHits given PeptideHits
        void computeCoverage(libcpp_vector[PeptideIdentification] pep_ids) nogil except +

        # Returns the date of the protein identification run
        DateTime getDateTime() nogil except +
        # Sets the date of the protein identification run
        void setDateTime(DateTime date) nogil except +
        # Sets the search engine type
        void setSearchEngine(String search_engine) nogil except +
        # Returns the type of search engine used
        String getSearchEngine() nogil except +
        # Sets the search engine version
        void setSearchEngineVersion(String search_engine_version) nogil except +
        # Returns the search engine version
        String getSearchEngineVersion() nogil except +
        # Sets the search parameters
        void setSearchParameters(SearchParameters search_parameters) nogil except +
        # Returns the search parameters
        SearchParameters getSearchParameters() nogil except +
        # Returns the identifier
        String getIdentifier() nogil except +
        # Sets the identifier
        void setIdentifier(String id_) nogil except +

        void setPrimaryMSRunPath(StringList& s) nogil except +
        void addPrimaryMSRunPath(StringList& s) nogil except +
        void getPrimaryMSRunPath(StringList& output) nogil except +

        void setPrimaryMSRunPath(StringList& s, bool raw) nogil except +
        void addPrimaryMSRunPath(StringList& s, bool raw) nogil except +
        void getPrimaryMSRunPath(StringList& output, bool raw) nogil except +

        # This causes as problem with circular dependencies when trying to use
        # ExperimentalSettings in MSExperiment
        # TODO: use addons if we really need this
        # void setPrimaryMSRunPath(StringList& s, MSExperiment& e) nogil except +

cdef extern from "<OpenMS/METADATA/ProteinIdentification.h>" namespace "OpenMS::ProteinIdentification":

    cdef enum PeakMassType:
        # wrap-attach:
        #    ProteinIdentification
        MONOISOTOPIC, AVERAGE, SIZE_OF_PEAKMASSTYPE

    cdef cppclass ProteinGroup:

      ProteinGroup()  nogil except +
      ProteinGroup(ProteinGroup)  nogil except +

      # Probability of this group
      double probability
      # Accessions of (indistinguishable) proteins that belong to the same group
      StringList accessions


    # Search parameters of the DB search
    cdef cppclass SearchParameters(MetaInfoInterface):
        # wrap-inherits:
        #    MetaInfoInterface

      SearchParameters()  nogil except +
      SearchParameters(SearchParameters) nogil except +

      String db            #< The used database
      String db_version            #< The database version
      String taxonomy            #< The taxonomy restriction
      String charges            #< The allowed charges for the search
      PeakMassType mass_type            #< Mass type of the peaks
      libcpp_vector[String] fixed_modifications            #< Used fixed modifications
      libcpp_vector[String] variable_modifications            #< Allowed variable modifications
      UInt missed_cleavages            #< The number of allowed missed cleavages
      double fragment_mass_tolerance            #< Mass tolerance of fragment ions (Dalton)
      bool fragment_mass_tolerance_ppm
      double precursor_mass_tolerance            #< Mass tolerance of precursor ions (Dalton)
      bool precursor_mass_tolerance_ppm
      DigestionEnzymeProtein digestion_enzyme            #< The enzyme for cleavage




from Types cimport *
from Peak1D cimport *
from String cimport *
from Software cimport *
from DateTime cimport *
from MetaInfoInterface cimport *

cdef extern from "<OpenMS/METADATA/DataProcessing.h>" namespace "OpenMS":

    cdef cppclass DataProcessing(MetaInfoInterface):
        # wrap-inherits:
        #     MetaInfoInterface

        DataProcessing()  nogil except +
        DataProcessing(DataProcessing) nogil except + # wrap-ignore

        void setProcessingActions(libcpp_set[ProcessingAction]) nogil except +
        libcpp_set[ProcessingAction] getProcessingActions() nogil except +

        Software getSoftware() nogil except +
        void setSoftware(Software s) nogil except +

        DateTime getCompletionTime()  nogil except +
        void setCompletionTime(DateTime t) nogil except +

    ctypedef shared_ptr[DataProcessing] DataProcessingPtr


cdef extern from "<OpenMS/METADATA/DataProcessing.h>" namespace "OpenMS::DataProcessing":

    cdef enum ProcessingAction:

        DATA_PROCESSING,                # General data processing (if no other term applies)
        CHARGE_DECONVOLUTION,           # Charge deconvolution
        DEISOTOPING,                    # Deisotoping
        SMOOTHING,                      # Smoothing of the signal to reduce noise
        CHARGE_CALCULATION,             # Determination of the peak charge
        PRECURSOR_RECALCULATION,        # Recalculation of precursor m/z
        BASELINE_REDUCTION,             # Baseline reduction
        PEAK_PICKING,                   # Peak picking (conversion from raw to peak data)
        ALIGNMENT,                      # Retention time alignment of different maps
        CALIBRATION,                    # Calibration of m/z positions
        NORMALIZATION,                  # Normalization of intensity values
        FILTERING,                      # Data filtering or extraction
        QUANTITATION,                   # Quantitation
        FEATURE_GROUPING,               # %Feature grouping
        IDENTIFICATION_MAPPING,         # %Identification mapping
        FORMAT_CONVERSION,              # General file format conversion (if no other term applies)
        CONVERSION_MZDATA,              # Conversion to mzData format
        CONVERSION_MZML,                # Conversion to mzML format
        CONVERSION_MZXML,               # Conversion to mzXML format
        CONVERSION_DTA,                 # Conversion to DTA format
        SIZE_OF_PROCESSINGACTION




from Types cimport *
from String cimport *

cdef extern from "<OpenMS/FORMAT/FileTypes.h>" namespace "OpenMS":

    cdef cppclass FileTypes:

        FileTypes() nogil except +
        FileTypes(FileTypes) nogil except +

        String typeToName(FileType t) nogil except +

        String typeToMZML(FileType t) nogil except +

        FileType nameToType(String name) nogil except +

cdef extern from "<OpenMS/FORMAT/FileTypes.h>" namespace "OpenMS::FileTypes":

    cdef enum FileType "OpenMS::FileTypes::Type":
    
          UNKNOWN,            # < Unknown file extension
          DTA,                # < DTA file (.dta)
          DTA2D,              # < DTA2D file (.dta2d)
          MZDATA,             # < MzData file (.mzData)
          MZXML,              # < MzXML file (.mzXML)
          FEATUREXML,         # < %OpenMS feature file (.featureXML)
          IDXML,              # < %OpenMS identification format (.idXML)
          CONSENSUSXML,       # < %OpenMS consensus map format (.consensusXML)
          MGF,                # < Mascot Generic Format (.mgf)
          INI,                # < %OpenMS parameters file (.ini)
          TOPPAS,             # < %OpenMS parameters file with workflow information (.toppas)
          TRANSFORMATIONXML,  # < Transformation description file (.trafoXML)
          MZML,               # < MzML file (.mzML)
          CACHEDMZML,         # < CachedMzML file (.cachedmzML)
          MS2,                # < MS2 file (.ms2)
          PEPXML,             # < TPP pepXML file (.pepXML)
          PROTXML,            # < TPP protXML file (.protXML)
          MZIDENTML,          # < mzIdentML (HUPO PSI AnalysisXML followup format) (.mzid)
          MZQUANTML,          # < mzQuantML (HUPO PSI AnalysisXML followup format) (.mzq)
          QCML,               # < qcML (will undergo standardisation maybe) (.qcml)
          GELML,              # < GelML (HUPO PSI format) (.gelML)
          TRAML,              # < TraML (HUPO PSI format) for transitions (.traML)
          MSP,                # < NIST spectra library file format (.msp)
          OMSSAXML,           # < OMSSA XML file format for peptide identifications (.xml)
          MASCOTXML,          # < Mascot XML file format for peptide identifications (.xml)
          PNG,                # < Portable Network Graphics (.png)
          XMASS,              # < XMass Analysis file (fid)
          TSV,                # < msInspect file (.tsv)
          PEPLIST,            # < specArray file (.peplist)
          HARDKLOER,          # < hardkloer file (.hardkloer)
          KROENIK,            # < kroenik file (.kroenik)
          FASTA,              # < FASTA file (.fasta)
          EDTA,               # < enhanced comma separated files (RT, m/z, Intensity, [meta])
          CSV,                # < general comma separated files format (might also be tab or space separated!!!), data should be regular, i.e. matrix form
          TXT,                # < any text format, which has only loose definition of what it actually contains -- thus it is usually hard to say where the file actually came from (e.g. PepNovo).
          OBO,                # < Controlled Vocabulary format
          HTML,               # < any HTML format
          XML,                # < any XML format
          ANALYSISXML,        # < analysisXML format
          XSD,                # < XSD schema format
          PSQ,                # < NCBI binary blast db
          MRM,                # < SpectraST MRM List
          SQMASS,             # < SqLite format for mass and chromatograms
          PQP,                # < OpenSWATH Peptide Query Parameter (PQP) SQLite DB
          OSW,                # < OpenSWATH OpenSWATH report (OSW) SQLite DB
          PSMS,               # < Percolator tab-delimited output (PSM level)
          PARAMXML,           # < internal format for writing and reading parameters (also used as part of CTD)
          SIZE_OF_TYPE        # < No file type. Simply stores the number of types




from libcpp.vector cimport vector as libcpp_vector
from Param cimport *
from String cimport *

cdef extern from "<OpenMS/DATASTRUCTURES/DefaultParamHandler.h>" namespace "OpenMS":

    cdef cppclass DefaultParamHandler:
        #wrap-ignore
        #no-pxd-import

        # DefaultParamHandler(const String & name) nogil except +
        # DefaultParamHandler(DefaultParamHandler & rhs) nogil except +
        # libcpp_vector[ String ] getSubsections() nogil except +

        void setParameters(Param &param)  nogil except +
        Param getParameters()  nogil except +
        Param getDefaults()  nogil except +
        String getName()  nogil except +
        void setName(const String&)  nogil except +



from libcpp.vector cimport vector as libcpp_vector
from String cimport *

from ProteinIdentification cimport *
from PeptideIdentification cimport *
from ModificationDefinitionsSet cimport *

cdef extern from "<OpenMS/FORMAT/XTandemXMLFile.h>" namespace "OpenMS":

    cdef cppclass XTandemXMLFile:

        XTandemXMLFile() nogil except +

        void load(String filename, ProteinIdentification & protein_identification,
                  libcpp_vector[PeptideIdentification] & id_data,
                  ModificationDefinitionsSet& mod_def_set) nogil except +




from Types cimport *
from MSExperiment cimport *
from ExperimentalSettings cimport *
from MSSpectrum cimport *
from MSChromatogram cimport *
from InterfaceDataStructures cimport *

cdef extern from "<OpenMS/KERNEL/OnDiscMSExperiment.h>" namespace "OpenMS":

    cdef cppclass OnDiscMSExperiment(ExperimentalSettings):

        OnDiscMSExperiment() nogil except +
        OnDiscMSExperiment(OnDiscMSExperiment &) nogil except +

        bool openFile(String filename) nogil except +
        bool openFile(String filename, bool skipLoadingMetaData) nogil except +
        Size getNrSpectra() nogil except +
        Size getNrChromatograms() nogil except +

        # COMMENT: only retrieves experiment meta data (no actual data in spectra/chromatograms)
        # COMMENT: useful for filtering by attributes to then retrieve data
        shared_ptr[const ExperimentalSettings] getExperimentalSettings() nogil except +
        shared_ptr[MSExperiment] getMetaData() nogil except +

        MSSpectrum getSpectrum(Size id) nogil except +
        MSSpectrum getSpectrumByNativeId(String id) nogil except +
        MSChromatogram getChromatogram(Size id) nogil except +
        MSChromatogram getChromatogramByNativeId(String id) nogil except +

        # TODO decide for 1.12 whether to include those ... 
        shared_ptr[Spectrum] getSpectrumById(int id_) nogil except +
        shared_ptr[Chromatogram] getChromatogramById(int id_) nogil except +

        void setSkipXMLChecks(bool skip) nogil except +




from String cimport *
from Types cimport *

cdef extern from "<OpenMS/DATASTRUCTURES/Date.h>" namespace "OpenMS":

    cdef cppclass Date:

        Date() nogil except +
        Date(Date) nogil except + # wrap-ignore

        void set(const String & date) nogil except +
        # void set(UInt month, UInt day, UInt year);

        Date today() nogil except +
        String get()  nogil except +
        # void get(UInt & month, UInt & day, UInt & year) nogil except +

        # Sets the undefined date: 00/00/0000
        void clear() nogil except +



from Types cimport *
from libcpp.vector cimport vector as libcpp_vector
from DefaultParamHandler cimport *
from ProgressLogger cimport *
from FASTAFile cimport *

cdef extern from "<OpenMS/ANALYSIS/ID/PeptideIndexing.h>" namespace "OpenMS":
    
    cdef cppclass PeptideIndexing(DefaultParamHandler) :
        # wrap-inherits:
        #  DefaultParamHandler
        PeptideIndexing() nogil except +
        PeptideIndexing(PeptideIndexing) nogil except + #wrap-ignore

        PeptideIndexing_ExitCodes run(libcpp_vector[ FASTAEntry ] & proteins,
                                      libcpp_vector[ ProteinIdentification ] & prot_ids,
                                      libcpp_vector[ PeptideIdentification ] & pep_ids) nogil except +

        String getDecoyString() nogil except +
        bool isPrefix() nogil except +

cdef extern from "<OpenMS/ANALYSIS/ID/PeptideIndexing.h>" namespace "OpenMS::PeptideIndexing":
    cdef enum PeptideIndexing_ExitCodes "OpenMS::PeptideIndexing::ExitCodes":
        #wrap-attach:
        #    PeptideIndexing
        EXECUTION_OK
        DATABASE_EMPTY
        PEPTIDE_IDS_EMPTY
        ILLEGAL_PARAMETERS
        UNEXPECTED_RESULT
        DECOYSTRING_EMPTY




from Types cimport *
from libcpp cimport bool
from libcpp.vector cimport vector as libcpp_vector
from DPosition cimport *
from DBoundingBox cimport *


# typedef DPosition<2> Point;
# typedef DBoundingBox<2> Rectangle;

cdef extern from "<OpenMS/COMPARISON/CLUSTERING/GridBasedCluster.h>" namespace "OpenMS":
    
    cdef cppclass GridBasedCluster "OpenMS::GridBasedCluster":

        GridBasedCluster(GridBasedCluster) nogil except + #wrap-ignore
        GridBasedCluster(DPosition2 centre,
            DBoundingBox2 bounding_box, 
            libcpp_vector[ int ] point_indices,
            int property_A, 
            libcpp_vector[ int ] properties_B) nogil except +
        GridBasedCluster(DPosition2 centre, 
           DBoundingBox2 bounding_box,
           libcpp_vector[ int ] point_indices) nogil except +

        DPosition2 getCentre() nogil except +
        DBoundingBox2 getBoundingBox() nogil except +
        libcpp_vector[ int ] getPoints() nogil except +
        int getPropertyA() nogil except +
        libcpp_vector[ int ] getPropertiesB() nogil except +

        # bool operator<(GridBasedCluster other) nogil except +
        # bool operator>(GridBasedCluster other) nogil except +
        # bool operator==(GridBasedCluster other) nogil except +





import unittest
import os

import pyopenms

class TestChromatogramExtractor(unittest.TestCase):

    def setUp(self):
        dirname = os.path.dirname(os.path.abspath(__file__))
        self.filename = os.path.join(dirname, "test.TraML").encode()
        self.filename_mzml = os.path.join(dirname, "test2.mzML").encode()

    def test_extractor(self):
        targeted = pyopenms.TargetedExperiment();
        tramlfile = pyopenms.TraMLFile();
        tramlfile.load(self.filename, targeted);

        exp = pyopenms.MSExperiment()
        pyopenms.MzMLFile().load(self.filename_mzml, exp)

        trafo = pyopenms.TransformationDescription()

        tmp_out = pyopenms.MSExperiment();
        extractor = pyopenms.ChromatogramExtractor()
        extractor.extractChromatograms(exp, tmp_out, targeted, 10, False, trafo, -1, b"tophat")

        # Basically test that the output is non-zero (e.g. the data is
        # correctly relayed to python)
        # The functionality is not tested here!
        self.assertEqual(len(tmp_out.getChromatograms()), len(targeted.getTransitions()))
        self.assertNotEqual(len(tmp_out.getChromatograms()), 0)
        self.assertEqual(tmp_out.getChromatograms()[0].size(), exp.size())
        self.assertNotEqual(tmp_out.getChromatograms()[0].size(), 0)
        self.assertNotEqual(tmp_out.getChromatograms()[0][0].getRT(), 0)
        self.assertNotEqual(tmp_out.getChromatograms()[0][0].getIntensity(), 0)

if __name__ == '__main__':
    unittest.main()


